#!/usr/bin/env python
# coding: utf-8

"""Notice: after you edit log templates,
it is NOT reccommended to add other log messages,
especially with new log templates.
On current implementation, lt_edit will destroy
the structures of log template manager."""

import sys

from . import config
from . import log_db
from . import lt_common


def export(ld):
    for ltline in ld.table:
        print(ltline)


def show_all(ld):
    print(ld.show_all_ltgroup())


def show_lt(ld):
    print(ld.show_all_lt())


def show_ltg(ld, gid):
    if gid is None:
        print(ld.show_all_ltgroup())
    else:
        print(ld.show_ltgroup(gid))


def show_sort(ld):
    buf = []
    for ltline in ld.iter_lt():
        buf.append(str(ltline))
    buf.sort()
    print("\n".join(buf))


def breakdown_ltid(ld, ltid, limit):
    d_args = {}
    for line in ld.iter_lines(ltid = ltid):
        for vid, arg in enumerate(line.var()):
            d_var = d_args.setdefault(vid, {})
            d_var[arg] = d_var.get(arg, 0) + 1

    buf = []
    
    buf.append("LTID {0}> {1}".format(ltid, str(ld.lt(ltid))))
    buf.append(" ".join(ld.lt(ltid).ltw))
    buf.append("")
    for vid, loc in enumerate(ld.lt(ltid).var_location()):
        buf.append("Variable {0} (word location : {1})".format(vid, loc))
        items = sorted(d_args[vid].items(), key=lambda x: x[1], reverse=True)
        var_variety = len(d_args[vid].keys())
        if var_variety > limit:
            for item in items[:limit]:
                buf.append("{0} : {1}".format(item[0], item[1]))
            buf.append("... {0} kinds of variable".format(var_variety))
        else:
            for item in items:
                buf.append("{0} : {1}".format(item[0], item[1]))
        buf.append("")
    return "\n".join(buf)


def _str_lt(ltid, ld):
    return "ltid {0} : {1}".format(ltid, str(ld.lt(ltid)))


#def _update_tpl(ld, old_ltw, new_ltw):
#    table = ld.ltm.ltgen._table
#    if table.exists(old_ltw):
#        tid = table.get_tid(old_ltw)
#        state = ld.ltm.ltgen.update_table(new_ltw, tid, False)
#    else:
#        raise ValueError("No existing tpl, failed")


def merge_ltid(ld, ltid1, ltid2):
    ld.init_ltmanager()
    sym = ld.ltm.sym
    print("merge following log templates...")
    print(_str_lt(ltid1, ld))
    print(_str_lt(ltid2, ld))
    print()

    ltw1 = ld.lt(ltid1).ltw
    cnt1 = ld.lt(ltid1).cnt
    l_s = ld.lt(ltid1).lts
    ltw2 = ld.lt(ltid2).ltw
    cnt2 = ld.lt(ltid2).cnt
    if not len(ltw1) == len(ltw2):
        sys.exit("log template length is different, failed")
    new_ltw = lt_common.merged_template(ltw1, ltw2, sym)

    ld.ltm.replace_lt(ltid1, new_ltw, l_s, cnt1 + cnt2)
    ld.ltm.remove_lt(ltid2)
    ld.db.update_log({"ltid" : ltid2}, {"ltid" : ltid1})

    ld.commit_db()

    print("> new log template : ltid {0}".format(ltid1))
    print(_str_lt(ltid1, ld))


def separate_ltid(ld, ltid, vid, value):

    def remake_lt(ld, ltid):
        # use iter_words rather than iter_lines
        # because iter_lines needs ltline registered in table
        # but ltline on new_ltid is now on construction in this function...
        new_lt = None
        cnt = 0
        for l_w in ld.db.iter_words(ltid = ltid):
            if new_lt is None:
                new_lt = l_w
            else:
                new_lt = lt_common.merged_template(new_lt, l_w, sym)
            cnt += 1
        print("result : " + str(new_lt))
        print()
        return new_lt, cnt

    ld.init_ltmanager()
    sym = ld.ltm.sym
    print("separate following log template...")
    print(_str_lt(ltid, ld))
    print("new log template if variable {0} is {1}".format(vid, value))
    print()

    l_lid = [lm.lid for lm in ld.iter_lines(ltid = ltid)
            if lm.var()[vid] == value]
    new_ltid = ld.lttable.next_ltid()    
    for lid in l_lid:
        ld.db.update_log({"lid" : lid}, {"ltid" : new_ltid})

    l_s = ld.lt(ltid).lts
    ltw1, cnt1 = remake_lt(ld, ltid)
    ld.ltm.replace_lt(ltid, ltw1, l_s, cnt1)

    ltw2, cnt2 = remake_lt(ld, new_ltid)
    ltline = ld.ltm.add_lt(ltw2, l_s, cnt2)
    assert ltline.ltid == new_ltid

    ld.commit_db()

    print("> new log templates : ltid {0}, ltid {1}".format(ltid, new_ltid))
    print(_str_lt(ltid, ld))
    print(_str_lt(new_ltid, ld))


def split_ltid(ld, ltid, vid):
    ld.init_ltmanager()
    sym = ld.ltm.sym
    print("split following log template...")
    print(_str_lt(ltid, ld))
    
    d_lid = {}
    for lm in ld.iter_lines(ltid = ltid):
        lid = lm.lid
        w = lm.var()[vid] 
        d_lid.setdefault(w, []).append(lid)
    
    print("variable {0}: {1}".format(vid, d_lid.keys()))
    if len(d_lid) > 10:
        print("variable candidate is too large... are you sure?")
        import pdb; pdb.set_trace()

    args = d_lid.keys()
    d_word_ltid = {}

    ltobj = ld.lt(ltid)
    l_s = ltobj.lts
    vloc = ltobj.var_location()[vid]

    # Update a template on existing ltid
    new_ltid = ltid
    new_word = args.pop(0)
    new_ltw = ltobj.ltw[:]; new_ltw[vloc] = new_word
    cnt = len(d_lid[new_word])
    ld.ltm.replace_lt(new_ltid, new_ltw, l_s, cnt)
    d_word_ltid[new_word] = new_ltid
    print("> new log templates : ltid {0}".format(new_ltid))
    print(_str_lt(new_ltid, ld))

    # Add templates
    while len(args) > 0:
        new_ltid = ld.lttable.next_ltid()   
        new_word = args.pop(0)
        new_ltw = ltobj.ltw[:]
        new_ltw[vloc] = new_word
        cnt = len(d_lid[new_word])
        new_ltobj = ld.ltm.add_lt(new_ltw, l_s, cnt)
        assert new_ltid == new_ltobj.ltid
        d_word_ltid[new_word] = new_ltid
        print("> new log templates : ltid {0}".format(new_ltid))
        print(_str_lt(new_ltid, ld))
    
    # Update log lines
    for word, l_lid in d_lid.items():
        temp_ltid = d_word_ltid[word]
        for lid in l_lid:
            ld.db.update_log({"lid" : lid}, {"ltid" : temp_ltid})

    ld.commit_db()


def fix_ltid(ld, ltid, l_vid):
    ld.init_ltmanager()
    sym = ld.ltm.sym
    print("make variable (with no variety) into description word...")
    print(_str_lt(ltid, ld))
    
    d_variety = {vid : set() for vid in l_vid}
    for lm in ld.iter_lines(ltid = ltid):
        for vid in l_vid:
            d_variety[vid].add(lm.var()[vid])

    d_fixed = {}
    for vid in l_vid:
        variety = d_variety[vid]
        assert len(variety) > 0
        if len(variety) == 1:
            print("confirmed that variable {0} is stable".format(vid))
            d_fixed[vid] = variety.pop()
            print("fixed word : {0}".format(d_fixed[vid]))
        else:
            print("variable {0} is not stable, ignored".format(vid))
            print(variety)

    ltobj = ld.lt(ltid)
    new_ltw = ltobj.ltw[:]
    for vid in d_fixed.keys():
        vloc = ltobj.var_location()[vid]
        new_ltw[vloc] = d_fixed[vid]
    l_s = ltobj.lts
    cnt = ltobj.cnt

    ld.ltm.replace_lt(ltid, new_ltw, l_s, cnt)
    ld.commit_db()
    print("> new log templates : ltid {0}".format(ltid))
    print(_str_lt(ltid, ld))


def free_ltid(ld, ltid, l_wid):
    ld.init_ltmanager()
    sym = ld.ltm.sym
    print("make description word into variable (with no variety)...")
    print(_str_lt(ltid, ld))
 
    ltobj = ld.lt(ltid)
    new_ltw = ltobj.ltw[:]
    for wid in l_wid:
        if ltobj.ltw[wid] == sym:
            print("wid {0} seems a variable, ignored")
        else:
            print("confirmed that wid {0} is description word ({1})".format(
                wid, ltobj.ltw[wid]))
            new_ltw[wid] = sym
    l_s = ltobj.lts
    cnt = ltobj.cnt

    ld.ltm.replace_lt(ltid, new_ltw, l_s, cnt)
    ld.commit_db()
    print("> new log templates : ltid {0}".format(ltid))
    print(_str_lt(ltid, ld))


def search_stable_variable(ld, th = 1):
    #ld.init_ltmanager()

    for ltobj in ld.iter_lt():
        ltid = ltobj.ltid
        d_args = {}
        for lm in ld.iter_lines(ltid = ltid):
            for vid, arg in enumerate(lm.var()):
                d_var = d_args.setdefault(vid, {})
                d_var[arg] = d_var.get(arg, 0) + 1
        for vid, loc in enumerate(ld.lt(ltid).var_location()):
            var_variety = len(d_args[vid].keys())
            if var_variety <= th:
                print("{0} {1}".format(ltobj.ltid, ltobj))
                print("variable {0} (word location {1}): {2}".format(
                        vid, loc, d_args[vid]))


def search_stable_vrule(ld, restr, dry = False):
    import re
    reobj = re.compile(restr)
    for ltobj in ld.iter_lt():
        ltid = ltobj.ltid
        d_args = {}
        for lm in ld.iter_lines(ltid = ltid):
            for vid, arg in enumerate(lm.var()):
                d_var = d_args.setdefault(vid, {})
                d_var[arg] = d_var.get(arg, 0) + 1
        for vid, loc in enumerate(ld.lt(ltid).var_location()):
            if len(d_args[vid]) == 1 and reobj.match(
                    list(d_args[vid].keys())[0]):
                print("{0} {1}".format(ltobj.ltid, ltobj))
                print("variable {0} (word location {1}): {2}".format(
                        vid, loc, list(d_args[vid].keys())[0]))
                if not dry:
                    fix_ltid(ld, ltid, [vid])

    
def search_desc_free(ld, restr, dry = False):
    import re
    reobj = re.compile(restr)
    for ltobj in ld.iter_lt():
        ltid = ltobj.ltid
        for wid, w in enumerate(ltobj.ltw):
            if w == ltobj.sym:
                pass
            elif reobj.match(w):
                print("{0} {1}".format(ltobj.ltid, ltobj))
                print("word location {0}".format(wid))
                if not dry:
                    free_ltid(ld, ltid, wid)


#if __name__ == "__main__":
#    
#    usage = """
#usage: {0} [options] args...
#args:
#  show : show all ltgroups
#  show-lt : show all log template without grouping
#  show-group LTGID : show log template group which has given LTGID
#  breakdown LTID : show variables appeared in log instances of given LTID
#  merge LTID1 LTID2 : merge log data with LTID1 and LTID2
#  separate LTID VID VALUE : make new log template with log data
#                            that have given variable VALUE in place VID
#    """.format(sys.argv[0]).strip()
#    
#    op = optparse.OptionParser(usage)
#    op.add_option("-c", "--config", action="store",
#            dest="conf", type="string", default=config.DEFAULT_CONFIG_NAME,
#            help="configuration file")
#    op.add_option("-l", "--limit", action="store",
#            dest="show_limit", type="int", default=5,
#            help="Limitation rows to show source log data")
#    (options, args) = op.parse_args()
#    if len(args) == 0:
#        sys.exit(usage)
#    mode = args.pop(0)
#    conf = config.open_config(options.conf)
#
#    ld = log_db.LogData(conf, edit = True)
#    if mode == "export":
#        export(ld)
#    elif mode == "show":
#        show_all(ld)
#    elif mode == "show-lt":
#        show_lt(ld)
#    elif mode == "show-group":
#        if len(args) == 0:
#            show_ltg(ld, None)
#        else:
#            show_ltg(ld, int(args[1]))
#    elif mode == "show-sort":
#        show_sort(ld)
#    elif mode == "breakdown":
#        if len(args) == 0:
#            sys.exit("give me ltid, following \"{0}\"".format(mode))
#        ltid = int(args[0])
#        print(breakdown_ltid(ld, ltid, options.show_limit))
#    elif mode == "merge":
#        if len(args) < 2:
#            sys.exit("give me 2 ltid, following \"{0}\"".format(mode))
#        ltid1 = int(args[0])
#        ltid2 = int(args[1])
#        sym = conf.get("log_template", "variable_symbol")
#        merge_ltid(ld, ltid1, ltid2, sym)
#    elif mode == "separate":
#        if len(args) < 3:
#            sys.exit("give me ltid, variable id and value, "
#                    "following \"{0}\"".format(mode))
#        ltid = int(args[0])
#        vid = int(args[1])
#        val = args[2]
#        sym = conf.get("log_template", "variable_symbol")
#        separate_ltid(ld, ltid, vid, val, sym)
#    elif mode == "split":
#        if len(args) < 2:
#            sys.exit("give me ltid and variable id, "
#                    "following \"{0}\"".format(mode))
#        ltid = int(args[0])
#        vid = int(args[1])
#        sym = conf.get("log_template", "variable_symbol")
#        split_ltid(ld, ltid, vid, sym)
#    elif mode == "fix":
#        if len(args) < 2:
#            sys.exit("give me ltid and variable id to fix, "
#                    "following \"{0}\"".format(mode))
#        ltid = int(args[0])
#        l_vid = [int(i) for i in args[1:]]
#        sym = conf.get("log_template", "variable_symbol")
#        fix_ltid(ld, ltid, l_vid, sym)
#    elif mode == "free":
#        if len(args) < 2:
#            sys.exit("give me ltid and word id to free, "
#                    "following \"{0}\"".format(mode))
#        ltid = int(args[0])
#        l_wid = [int(i) for i in args[1:]]
#        sym = conf.get("log_template", "variable_symbol")
#        free_ltid(ld, ltid, l_wid, sym)
#    elif mode == "search-stable":
#        if len(args) >= 1:
#            th = int(args[0])
#        else:
#            th = 1
#        search_stable_variable(ld, th)




#!/usr/bin/env python
# coding: utf-8

"""
Note:
    Configuration defined earlier is prior.
"""


import sys
import os
import re
import logging
import configparser
from collections import defaultdict

from . import common
from . import config
from . import strutil
from . import logparser
from . import log_db

_logger = logging.getLogger(__package__)
DEFAULT_LABEL_CONF = "/".join((os.path.dirname(__file__),
                               "data/lt_label.conf.sample"))


class LTLabel():

    group_header = "group_"
    label_header = "label_"

    def __init__(self, conf_fn, default_label = None, default_group = None):
        self.conf = configparser.ConfigParser()
        self.conf.read(conf_fn)
        self.default_label = default_label
        self.default_group = default_group

        self.groups = []
        self.labels = []
        for sec in self.conf.sections():
            if sec[:len(self.group_header)] == self.group_header:
                self.groups.append(sec[len(self.group_header):])
            elif sec[:len(self.label_header)] == self.label_header:
                self.labels.append(sec[len(self.label_header):])

        self.d_group = {} # key : group, val : [label, ...]
        self.d_rgroup = {} # key : label, val : [group, ...]
        for group in self.groups:
            section = self.group_header + group
            for label in config.gettuple(self.conf, section, "members"):
                self.d_group.setdefault(group, []).append(label)
                self.d_rgroup.setdefault(label, []).append(group)
        self.rules = [] # [(label, (re_matchobj, ...)), ...]
        for label in self.labels:
            section = self.label_header + label
            for rulename in config.gettuple(self.conf, section, "rules"):
                l_re = []
                l_restr = config.gettuple(self.conf, section, rulename)
                for re_str in l_restr:
                    if rulename[0] == "i":
                        re_obj = re.compile(re_str, re.IGNORECASE)
                    elif rulename[0] == "c":
                        re_obj = re.compile(re_str, )
                    else:
                        raise ValueError(
                                "lt_label rulename invalid ({0})".format(
                                    rulename))
                    l_re.append(re_obj)
                self.rules.append((label, tuple(l_re)))

            #    l_word = config.gettuple(self.conf, section, rulename + "_word")
            #    l_rule = confoig.gettuple(self.conf, section, rulename + "_rule")
            #    assert len(l_word) == len(l_rule)
            #    self.rules.append((label, l_word, l_rule))

    #def _test_rule(self, ltline, l_word, l_rule):
    #    for word, rule in zip(l_word, l_rule):
    #        if rule == "equal":
    #            if word in ltline.ltw:
    #                # satisfied
    #                pass
    #            else:
    #                return False
    #        if rule == "equal_ord":
    #            if word.lower() in [w.lower() for w in ltline.ltw]:
    #                # satisfied
    #                pass
    #            else:
    #                return False
    #        elif rule == "in":
    #            for w in ltline.ltw:
    #                if word in w:
    #                    # satisfied
    #                    break
    #            else:
    #                return False
    #        elif rule == "in_ord":
    #            for w in ltline.ltw:
    #                if word.lower() in w.lower():
    #                    # satisfied
    #                    break
    #            else:
    #                return False
    #        else:
    #            raise ValueError("Invalid rule name")
    #    else:
    #        return True

    def _test_rule(self, ltline, t_re):
        for reobj in t_re:
            for word in ltline.ltw:
                if reobj.match(word):
                    break
                else:
                    pass
            else:
                # no word matches
                return False
        else:
            # all reobj passes
            return True

    def get_lt_label(self, ltline):
        for label, t_re in self.rules:
            if self._test_rule(ltline, t_re):
                return label
        else:
            return self.default_label

    def get_lt_group(self, ltline):
        label = self.get_lt_label(ltline)
        return self.get_group(label)

    def get_ltg_label(self, ltgid, l_ltline):
        d_score = {} # key : ruleid, value : score
        for ltline in l_ltline:
            for rid, t_rule in enumerate(self.rules):
                label, t_re = t_rule
                if self._test_rule(ltline, t_re):
                    d_score[rid] = d_score.get(rid, 0) + 1
        l_cand = []
        max_score = 0
        for rid, score in d_score.items():
            if score > max_score:
                max_score = score
                l_cand = [rid]
            elif score == max_score:
                l_cand.append(rid)
        if len(l_cand) > 1:
            _logger.debug("multiple label for ltgid {0} : {1}".format(\
                ltgid, [self.rules[rid][0] for rid in l_cand]))
            return self.rules[l_cand[0]][0]
        elif len(l_cand) == 1:
            return self.rules[l_cand[0]][0]
        else:
            return self.default_label

    def get_ltg_group(self, ltgid, l_ltline):
        label = self.get_ltg_label(ltgid, l_ltline)
        return self.get_group(label)

    def get_gid_label(self, gid, gid_name, ld):
        if gid_name == "ltid":
            return self.get_lt_label(ld.lt(gid))
        elif gid_name == "ltgid":
            return self.get_ltg_label(gid, ld.ltg_members(gid))

    def get_gid_group(self, gid, gid_name, ld):
        if gid_name == "ltid":
            return self.get_lt_group(ld.lt(gid))
        elif gid_name == "ltgid":
            return self.get_ltg_group(gid, ld.ltg_members(gid))

    def get_group(self, label):
        if label is None:
            return self.default_group
        else:
            if label in self.d_rgroup:
                group = self.d_rgroup[label]
                assert len(group) == 1
                return group[0]
            else:
                return self.default_group


def init_ltlabel(conf):
    ltconf_path = conf.get("visual", "ltlabel")
    if ltconf_path == "":
        ltconf_path = DEFAULT_LABEL_CONF
    default_label = conf.get("visual", "ltlabel_default_label")
    default_group = conf.get("visual", "ltlabel_default_group")
    return LTLabel(ltconf_path, default_label = default_label,
                   default_group = default_group)


def count_ltlabel(conf):
    ld = log_db.LogData(conf)
    ll = init_ltlabel(conf)
    default_label = conf.get("visual", "ltlabel_default_label")
    default_group = conf.get("visual", "ltlabel_default_group")

    d_lt_group = defaultdict(int)
    d_lt_label = defaultdict(int)
    d_line_group = defaultdict(int)
    d_line_label = defaultdict(int)
    for ltgid in ld.iter_ltgid():
        l_gid = ld.ltg_members(ltgid)
        l_lt = ld.ltg_members(ltgid)
        label = ll.get_ltg_label(ltgid, l_lt)
        group = ll.get_group(label)

        if label is None:
            label = default_label
            group = default_group

        d_lt_group[group] += 1
        d_lt_label[label] += 1

        cnt_line = sum(lt.cnt for lt in l_lt)
        d_line_group[group] += cnt_line
        d_line_label[label] += cnt_line

    print("all templates : {0}".format(sum(d_lt_group.values())))
    print("all lines : {0}".format(sum(d_line_group.values())))
    print()

    for group, l_label in ll.d_group.items():
        if group in d_lt_group:
            cnt_group = d_lt_group.pop(group)
            lines_group = d_line_group.pop(group)
        else:
            cnt_group = 0; lines_group = 0
        #cnt_group = d_lt_group[group]
        #lines_group = d_line_group[group]
        print("group {0} : {1} templates, {2} lines".format(group,
                cnt_group, lines_group))

        for label in l_label:
            if label in d_lt_label:
                cnt_label = d_lt_label.pop(label)
                lines_label = d_line_label.pop(label)
            else:
                cnt_label = 0; lines_label = 0
            #cnt_label = d_lt_label[label]
            #lines_label = d_line_label[label]
            print("  label {0} : {1} templates, {2} lines".format(label,
                    cnt_label, lines_label))
        print()
    
    print(d_line_group); print(d_line_label)


def list_ltlabel(conf):

    def output(ld, ltgid, label, group):
        if label is None:
            label = str(label)
        return " ".join((group, label, ld.show_ltgroup(ltgid)))

    ld = log_db.LogData(conf)
    ll = init_ltlabel(conf)
    
    d_buf = {}
    buf_none = []
    for ltgid in ld.iter_ltgid():
        l_gid = ld.ltg_members(ltgid)
        if len(l_gid) == 1:
            #label = ll.get_lt_label(ltgid, ld.ltg_members(ltgid))
            label = ll.get_ltg_label(ltgid, ld.ltg_members(ltgid))
            group = ll.get_group(label)
        else:
            label = ll.get_ltg_label(ltgid, ld.ltg_members(ltgid))
            group = ll.get_group(label)
        if label is None:
            buf_none.append(output(ld, ltgid, str(label), group))
        else:
            if group is None:
                raise SyntaxError("bad config for label {0}: "
                                  "No group found".format(label))
            d_buf.setdefault(label, []).append(output(ld, ltgid, label, group))
    for k, buf in sorted(d_buf.items()):
        print("\n".join(buf))
        print()
    print("\n".join(buf_none))


def count_event_label(conf):
    import log2event
    ld = log_db.LogData(conf)
    ll = init_ltlabel(conf)    
    d_label = defaultdict(int)
    d_group = defaultdict(int)
    
    dirname = conf.get("dag", "event_dir")
    for fp in common.rep_dir(dirname):
        fn = fp.split("/")[-1]
        edict, evmap = log2event.load_edict(fp)
        for eid, l_dt in edict.items():
            gid = evmap.info(eid).gid
            l_lt = ld.ltg_members(gid)
            label = ll.get_ltg_label(gid, l_lt)
            group = ll.get_group(label)

            d_label[label] += len(l_dt)
            d_group[group] += len(l_dt)

    print("all lines : {0}".format(sum(d_group.values())))
    print()
    
    for group, l_label in ll.d_group.items():
        if group in d_group:
            cnt_group = d_group.pop(group)
        else:
            cnt_group = 0
        print("group {0}: {1} lines".format(group, cnt_group))
        for label in l_label:
            if label in d_label:
                cnt_label = d_label.pop(label)
            else:
                cnt_label = 0
            print("  label {0}: {1} lines".format(label, cnt_label))
        print()


def count_edge_label(conf):
    ll = init_ltlabel(conf)
    import pcresult
    d_cnt_label = defaultdict(int)
    d_cnt_group = defaultdict(int)
    src_dir = conf.get("dag", "output_dir")
    for fp in common.rep_dir(src_dir):
        _logger.info("count_edge_label processing {0}".format(fp))
        r = pcresult.PCOutput(conf).load(fp)
        for edge in r.graph.edges():
            for eid in edge:
                gid = r.evmap.info(eid).gid
                label = r._label_ltg(gid)
                d_cnt_label[label] += 1
                group = r._label_group_ltg(gid)
                d_cnt_group[group] += 1

    for group, l_label in ll.d_group.items():
        cnt_group = d_cnt_group[group]
        print("group {0}: {1} nodes".format(group, cnt_group))
        for label in l_label:
            cnt_label = d_cnt_label[label]
            print("  label {0}: {1} nodes".format(label, cnt_label))
        print


def count_edge_label_detail(conf):
    ll = init_ltlabel(conf)
    ld = log_db.LogData(conf)
    import pcresult
    d_group = defaultdict(int)
    d_group_directed = defaultdict(int)
    d_group_intype = defaultdict(int)
    d_group_intype_directed = defaultdict(int)
    d_group_mean = defaultdict(int)
    d_group_mean_directed = defaultdict(int)
    import edge_filter
    ef = edge_filter.EdgeFilter(conf)

    src_dir = conf.get("dag", "output_dir")
    for fp in common.rep_dir(src_dir):
        _logger.info("count_edge_label_detail processing {0}".format(fp))
        r = pcresult.PCOutput(conf).load(fp)
        dedges, udedges = r._separate_edges()
        for edge in dedges:
            cedge = [r.evmap.info(eid) for eid in edge]
            fflag = ef.isfiltered(cedge)
            l_group = [r._label_group_ltg(r.evmap.info(eid).gid)
                    for eid in edge]
            iflag = (l_group[0] == l_group[1])

            for group in l_group:
                d_group[group] += 1
                d_group_directed[group] += 1
                if iflag:
                    d_group_intype[group] += 1
                    d_group_intype_directed[group] += 1
                if not fflag:
                    d_group_mean[group] += 1
                    d_group_mean_directed[group] += 1

        for edge in udedges:
            cedge = [r.evmap.info(eid) for eid in edge]
            fflag = ef.isfiltered(cedge)
            l_group = [r._label_group_ltg(r.evmap.info(eid).gid)
                    for eid in edge]
            iflag = (l_group[0] == l_group[1])

            for group in l_group:
                d_group[group] += 1
                if iflag:
                    d_group_intype[group] += 1
                if not fflag:
                    d_group_mean[group] += 1

    table = [["key", "all", "directed", "intype", "intype_directed",
            "important", "important_directed"]]
    for key in d_group.keys():
        temp = [key]
        temp.append(d_group[key])
        temp.append(d_group_directed[key])
        temp.append(d_group_intype[key])
        temp.append(d_group_intype_directed[key])
        temp.append(d_group_mean[key])
        temp.append(d_group_mean_directed[key])
        table.append(temp)

    table.append(["total", sum(d_group.values()),
            sum(d_group_directed.values()),
            sum(d_group_intype.values()),
            sum(d_group_intype_directed.values()),
            sum(d_group_mean.values()),
            sum(d_group_mean_directed.values())])

    print(common.cli_table(table))


def count_edge_label_extype(conf):
    ll = init_ltlabel(conf)
    ld = log_db.LogData(conf)
    import pcresult
    s_keys = set()
    d_extype = defaultdict(int)

    src_dir = conf.get("dag", "output_dir")
    for fp in common.rep_dir(src_dir):
        r = pcresult.PCOutput(conf).load(fp)
        dedges, udedges = r._separate_edges()
        for edge in dedges + udedges:
            l_group = [r._label_group_ltg(r.evmap.info(eid).gid)
                    for eid in edge]
            for group in l_group:
                s_keys.add(group)
            if l_group[0] == l_group[1]:
                d_extype[tuple(l_group)] += 1
            else:
                d_extype[(l_group[0], l_group[1])] += 1
                d_extype[(l_group[1], l_group[0])] += 1

    table = []
    table.append(["group"] + list(s_keys))
    for key1 in s_keys:
        buf = [key1]
        for key2 in s_keys:
            cnt = d_extype[(key1, key2)]
            buf.append(cnt)
        table.append(buf)

    print(common.cli_table(table))


def search_edge_label_extype(conf, label1, label2):
    ll = init_ltlabel(conf)
    ld = log_db.LogData(conf)
    import pcresult

    src_dir = conf.get("dag", "output_dir")
    for fp in common.rep_dir(src_dir):
        r = pcresult.PCOutput(conf).load(fp)
        rflag = False
        dedges, udedges = r._separate_edges()
        for edge in dedges + udedges:
            l_group = [r._label_group_ltg(r.evmap.info(eid).gid)
                    for eid in edge]
            if (l_group[0] == label1 and l_group[1] == label2) or \
                    (l_group[1] == label1 and l_group[0] == label2):
                if not rflag:
                    print("# {0}".format(r.filename))
                    rflag = True
                r._print_edge(edge, False)
                r._print_edge_lt(edge)


if __name__ == "__main__":
    import optparse
    usage = "usage: {0} [options] mode".format(sys.argv[0])
    op = optparse.OptionParser(usage)
    op.add_option("-c", "--config", action="store",
            dest="conf", type="string", default=config.DEFAULT_CONFIG_NAME,
            help="configuration file path")
    options, args = op.parse_args()
    conf = config.open_config(options._conf)
    config.set_common_logging(conf, _logger, [])

    if len(args) == 0:
        sys.exit(usage)
    mode = args.pop(0)
    if mode == "list":
        list_ltlabel(conf)
    elif mode == "count":
        count_ltlabel(conf)
    elif mode == "event":
        count_event_label(conf)
    elif mode == "edge":
        count_edge_label(conf)
    elif mode == "edge-detail":
        count_edge_label_detail(conf)
    elif mode == "edge-extype":
        count_edge_label_extype(conf)
    elif mode == "edge-search-extype":
        if len(args) < 2:
            sys.exit("give me 2 label names")
        search_edge_label_extype(conf, args[0], args[1])
    else:
        raise NotImplementedError






from collections import deque
import logging

from amulog import lt_common

_logger = logging.getLogger(__package__)


class LTSearch(object):

    def __init__(self):
        self._l_lt = []

    def add(self, ltid, l_w):
        self._l_lt.append((ltid, l_w))

    def search(self, l_w):
        for ltid, tmp_l_w in self._l_lt:
            if len(l_w) == len(tmp_l_w):
                for w1, w2 in zip(l_w, tmp_l_w):
                    if w1 == w2:
                        pass
                    elif w2 == lt_common.REPLACER:
                        pass
                    else:
                        break
                else:
                    return ltid
        else:
            return None

    def shuffle(self):
        import random
        random.shuffle(self._l_lt)


class LTSearchTree(LTSearch):
    # Search tree for un-incremental lt generation algorithms

    def __init__(self):
        self.root = self._new_node()

    def __str__(self):
        l_buf = []

        def print_children(point, depth, l_sparent):
            word = point.word
            if word is None:
                word = lt_common.REPLACER

            cnt = point.child_num()
            if cnt == 1:
                l_sparent.append(word)
            else:
                l_sparent.append(word)
                buf = "-" * (depth - len(l_sparent) + 1) + \
                      " {0}".format(" ".join(l_sparent))
                if point.end is not None:
                    buf += "  <-- ltid {0}".format(point.end)
                l_buf.append(buf)
                l_sparent = []

            for word in point.windex.keys():
                print_children(point.windex[word], depth + 1, l_sparent[:])
            if point.wild is not None:
                print_children(point.wild, depth + 1, l_sparent[:])

        point = self.root
        l_buf.append("<head of log template search tree>")
        for w in point.windex.keys():
            print_children(point.windex[w], 0, [])
        if point.wild is not None:
            print_children(point.wild, 0, [])
        return "\n".join(l_buf)

    @staticmethod
    def _new_node(parent=None, word=None):
        return LTSearchTreeNode(parent, word)

    def add(self, ltid, l_w):
        point = self.root
        for w in l_w:
            if w == lt_common.REPLACER:
                if point.wild is None:
                    point.wild = self._new_node(point, w)
                point = point.wild
            else:
                if not w in point.windex:
                    point.windex[w] = self._new_node(point, w)
                point = point.windex[w]
        else:
            point.set_ltid(ltid)

    def _trace(self, ltwords):
        _logger.debug("tracing message {0}".format(ltwords))
        point = self.root
        tmp_ltwords = deque(ltwords)
        check_points = []
        # points with 2 candidates to go, for example "word" and "**"
        # [(node, len(left_words)), ...]
        # use as stack
        while True:
            w = tmp_ltwords.popleft()
            if w == lt_common.REPLACER:
                # w is sym (This means the input is not a message but a template)
                # go wild node
                if point.wild is None:
                    return None
                else:
                    point = point.wild
            elif w in point.windex:
                # w is in windex
                # go the word node
                # also go to wild node, when check_points poped
                if point.wild is not None:
                    check_points.append((point, deque(tmp_ltwords)))
                    _logger.debug("# add checkpoint ({0} (+{1}))".format(point, len(tmp_ltwords) + 1))
                _logger.debug("{0} (+{1}) -> goto {2}".format(
                    point, len(tmp_ltwords) + 1, w))
                point = point.windex[w]
            elif point.wild is not None:
                # w is not in windex, but have wild node
                _logger.debug("{0} (+{1}) -> goto {2}".format(
                    point, len(tmp_ltwords) + 1, lt_common.REPLACER))
                point = point.wild
            else:
                # no template to process w, go back or end
                _logger.debug("{0} (+{1}) : no available children".format(point, len(tmp_ltwords) + 1))
                if len(check_points) == 0:
                    _logger.debug("template not found")
                    return None
                else:
                    p, tmp_ltwords = check_points.pop()
                    # tmp_ltwords = deque(ltwords[-left_wlen:]) if left_wlen > 0 else []
                    #    # +1 : for one **(wild) node
                    point = p.wild
                    _logger.debug("go back to a stacked branch : {0} (+{1})".format(point, len(tmp_ltwords)))
                    _logger.debug("remaining words : {0}".format(tmp_ltwords))

            while len(tmp_ltwords) == 0:
                if point.end is None:
                    _logger.debug("{0} (+{1}) : no template in this node".format(point, len(tmp_ltwords)))
                    if len(check_points) == 0:
                        _logger.debug("all ends are empty(no templates)")
                        return None
                    else:
                        p, tmp_ltwords = check_points.pop()
                        # tmp_ltwords = deque(ltwords[-left_wlen:]) if left_wlen > 0 else []
                        point = p.wild
                        _logger.debug("go back to a stacked branch : {0} (+{1})".format(point, len(tmp_ltwords)))
                        _logger.debug("remaining words : {0}".format(tmp_ltwords))
                else:
                    _logger.debug("done (tpl: {0}, id: {1})".format(point, point.end))
                    return point

    def remove(self, ltid, l_w):
        point = self._trace(l_w)
        if point is None:
            _logger.warning(
                "LTSearchTree : Failed to remove ltid {0}".format(ltid))
        point.remove_ltid(ltid)
        while point.unnecessary():
            w = point.word
            point = point.parent
            if w is None:
                point.wild = None
            else:
                point.wdict.pop(w)
        else:
            if self.root is None:
                self.root = self._new_node()

    def search(self, l_w):
        node = self._trace(l_w)
        if node is None:
            return None
        else:
            return node.get_ltid()

    def shuffle(self):
        pass


class LTSearchTreeNode:

    def __init__(self, parent, word):
        self.windex = {}
        self.wild = None
        self.end = None
        self.parent = parent  # for reverse search to remove
        self.word = word

    def __str__(self):
        ret = []
        p = self
        while p.parent is not None:
            if p.word is None:
                ret.append("<root>")
            else:
                ret.append(p.word)
            p = p.parent
        return " ".join(reversed(ret))

    def child(self, word=None):
        if word is None:
            # wildcard
            return self.wild
        elif word in self.windex:
            return self.windex[word]
        else:
            return None

    def child_num(self):
        cnt = len(self.windex.keys())
        if self.wild is not None:
            cnt += 1
        return cnt

    def current_point(self):
        buf = []
        point = self
        while point.parent is not None:
            buf = [point.word] + buf
            point = point.parent
        print(" ".join(buf))

    def set_ltid(self, ltid):
        self.end = ltid

    def remove_ltid(self, ltid):
        assert self.end == ltid
        self.end = None

    def get_ltid(self):
        return self.end

    def unnecessary(self):
        return (len(self.windex) == 0) and \
               (self.wild is None) and \
               (self.end is None)


def init_searcher(name):
    if name == "tree":
        return LTSearchTree()
    elif name == "table":
        return LTSearch()
    else:
        raise NotImplementedError



#!/usr/bin/env python
# coding: utf-8

import re
import logging
import ipaddress
import configparser

from amulog import lt_common
from amulog import host_alias

_logger = logging.getLogger(__package__)


class VariableRegex:

    def __init__(self, conf, fn=None, label_unknown="unknown"):
        self._ext = {}
        self._re = {}
        self.label_unknown = label_unknown

        self._ha = host_alias.init_hostalias(conf)
        if fn is not None and fn.strip() != "":
            self._load_rule(fn)

    def _load_rule(self, fn):

        def gettuple(conf, sec, opt):
            s = conf[sec][opt]
            if s.strip() == "":
                return tuple()
            else:
                return tuple([w.strip() for w in s.split(",")])

        vre_conf = configparser.ConfigParser()
        loaded = vre_conf.read(fn)
        if len(loaded) == 0:
            mes = "VariableRegex config {0} is empty".format(fn)
            _logger.warning(mes)

        t_ext = gettuple(vre_conf, "ext", "rules")
        for rule in t_ext:
            self._ext[rule] = getattr(self, vre_conf["ext"][rule].strip())

        t_re_rules = gettuple(vre_conf, "re", "rules")
        for rule in t_re_rules:
            tmp = []
            t_re = gettuple(vre_conf, "re", rule)
            for restr in t_re:
                tmp.append(re.compile(restr))
            self._re[rule] = tmp

    def match(self, word):
        for key, func in self._ext.items():
            if func(word):
                return True

        for key, l_reobj in self._re.items():
            for reobj in l_reobj:
                if reobj.match(word):
                    return True
        return False

    def label(self, word):
        for key, func in self._ext.items():
            if func(word):
                return key

        for key, l_reobj in self._re.items():
            for reobj in l_reobj:
                if reobj.match(word):
                    return key

        return self.label_unknown

    @staticmethod
    def label_ipaddr(word):
        try:
            ret = ipaddress.ip_address(str(word))
            if isinstance(ret, ipaddress.IPv4Address):
                return "IPv4ADDR"
            elif isinstance(ret, ipaddress.IPv6Address):
                return "IPv6ADDR"
            else:
                raise TypeError("ip_address returns unknown type? {0}".format(
                    str(ret)))
        except ValueError:
            return None

    @staticmethod
    def label_ipnetwork(word):
        try:
            ret = ipaddress.ip_network(str(word), strict=False)
            if isinstance(ret, ipaddress.IPv4Network):
                return "IPv4NET"
            elif isinstance(ret, ipaddress.IPv6Network):
                return "IPv6NET"
            else:
                raise TypeError("ip_address returns unknown type? {0}".format(
                    str(ret)))
        except ValueError:
            return None

    def label_host(self, word):
        if self._ha.isknown(word):
            return "HOST"
        else:
            return None


class LTGenRegularExpression(lt_common.LTGenStateless):

    def __init__(self, table, vreobj):
        super(LTGenRegularExpression, self).__init__(table)
        self._vre = vreobj

    def generate_tpl(self, pline):
        l_w = pline["words"]
        tpl = []
        for w in l_w:
            if self._vre.match(w):
                tpl.append(lt_common.REPLACER)
            else:
                tpl.append(w)
        return tpl


def init_ltgen_regex(conf, table, **_):
    fn = conf.get("log_template_re", "variable_rule")
    vreobj = VariableRegex(conf, fn)
    return LTGenRegularExpression(table, vreobj)



#!/usr/bin/env python
# coding: utf-8


ESC_LETTER = "*@"  # including back slash


def split_igesc(string, spl):
    l_esc = ["\\" + w for w in "\\" + ESC_LETTER]
    temp_str = string
    spl_len = len(spl)
    temp_ww = []
    ret = []
    while len(temp_str) >= spl_len:
        if temp_str[0:2] in l_esc:
            temp_ww.append(temp_str[:2])
            temp_str = temp_str[2:]
        elif fmatch(temp_str, spl):
            ret.append("".join(temp_ww))
            temp_ww = []
            temp_str = temp_str[len(spl):]
        else:
            temp_ww.append(temp_str[0])
            temp_str = temp_str[1:]
    else:
        ret.append("".join(temp_ww) + temp_str)
    return ret


def add_esc(buf):
    for w in "\\" + ESC_LETTER:
        buf = buf.replace(w, "\\" + w)
    return buf


def restore_esc(buf):
    for w in ESC_LETTER + "\\":
        buf = buf.replace("\\" + w, w)
    return buf


def fmatch(string, match):
    """Test the forward part of string matches another string or not.
    """
    assert len(string) >= len(match)
    return string[:len(match)] == match



#!/usr/bin/env python
# coding: utf-8

"""
Interface to use some functions about Log DB from CLI.
"""

import sys
import datetime
import logging
import argparse

from . import common
from . import config

_logger = logging.getLogger(__package__)
SUBLIB = ["eval", "crf"]


def get_targets_arg(ns):
    if ns.recur:
        targets = common.recur_dir(ns.files)
    else:
        targets = common.rep_dir(ns.files)
    return targets


def get_targets_conf(conf):
    l_path = config.getlist(conf, "general", "src_path")
    if conf.getboolean("general", "src_recur"):
        return common.recur_dir(l_path)
    else:
        return common.rep_dir(l_path)


def get_targets(ns, conf):
    if ns is None or len(ns.files) == 0:
        return get_targets_conf(conf)
    else:
        return get_targets_arg(ns)


def is_online(conf):
    from .alg import is_online as alg_is_online
    mode = conf["log_template"]["processing_mode"]
    lt_methods = config.getlist(conf, "log_template", "lt_methods")
    return alg_is_online(mode, lt_methods)


def data_from_db(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    dirname = ns.dirname
    if ns.incr:
        method = "incremental"
    else:
        method = "commit"
    reset = ns.reset

    from . import log_db
    log_db.data_from_db(conf, dirname, method, reset)


def data_from_data(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    dirname = ns.dirname
    targets = get_targets(ns, conf)
    if ns.incr:
        method = "incremental"
    else:
        method = "commit"
    reset = ns.reset

    from . import log_db
    log_db.data_from_data(conf, targets, dirname, method, reset)


def db_make(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    targets = get_targets(ns, conf)
    dry = ns.dry
    from . import log_db

    timer = common.Timer("db-make", output=_logger)
    timer.start()
    if is_online(conf):
        log_db.process_files_online(conf, targets, True, dry=dry)
    else:
        log_db.process_files_offline(conf, targets, dry=dry)
    timer.stop()


def db_add(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    targets = get_targets_arg(ns)
    from . import log_db

    if not is_online(conf):
        msg = "db_add is not available with offline template generation"
        sys.exit(msg)

    timer = common.Timer("db-add", output=_logger)
    timer.start()
    log_db.process_files_online(conf, targets, False)
    timer.stop()


def db_anonymize(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db

    timer = common.Timer("db-anonymize", output=_logger)
    timer.start()
    log_db.anonymize(conf)
    timer.stop()


def reload_area(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db
    log_db.reload_area(conf)


def show_db_info(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db

    log_db.info(conf)


def show_lt(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db

    simple = ns.simple
    log_db.show_lt(conf, simple=simple)


def show_ltg(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db

    kwargs = {}
    if ns.group is not None:
        kwargs["group"] = ns.group
    log_db.show_lt(conf, **kwargs)


def show_lt_import(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db

    external = ns.external
    ld = log_db.LogData(conf)
    for ltobj in ld.iter_lt():
        if external:
            # unsegmented, with escape
            print(ltobj.restore_message(None, esc=True))
        else:
            # segmented, with escape
            print(" ".join(ltobj.ltw))


#def show_lt_import_exception(ns):
#    conf = config.open_config(ns.conf_path)
#    lv = logging.DEBUG if ns.debug else logging.INFO
#    config.set_common_logging(conf, logger=_logger, lv=lv)
#
#    form = ns.format
#    assert form in ("log", "message")
#    targets = get_targets_arg(ns)
#    from . import lt_import
#
#    lt_import.search_exception(conf, targets, form)


def show_lt_words(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db

    d = log_db.agg_words(conf, target="all")
    print(common.cli_table(sorted(d.items(), key=lambda x: x[1],
                                  reverse=True)))


def show_lt_descriptions(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db

    d = log_db.agg_words(conf, target="description")
    print(common.cli_table(sorted(d.items(), key=lambda x: x[1],
                                  reverse=True)))


def show_lt_variables(ns):
    import re

    def repl_var(d):
        reobj = re.compile(r"[0-9]+")
        keys = list(d.keys())
        for k in keys:
            new_k = reobj.sub(r"\d", k)
            if k == new_k:
                pass
            else:
                d[new_k] += d.pop(k)

    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db

    d = log_db.agg_words(conf, target="variable")
    if ns.repld:
        repl_var(d)
    print(common.cli_table(sorted(d.items(), key=lambda x: x[1],
                                  reverse=True)))


def show_lt_breakdown(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db
    from . import lt_tool

    ld = log_db.LogData(conf)
    ltid = ns.ltid
    limit = ns.lines
    ret = lt_tool.breakdown_ltid(ld, ltid, limit)
    print(ret)


def show_lt_vstable(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db
    from . import lt_tool

    ld = log_db.LogData(conf)
    lt_tool.search_stable_variable(ld, th=1)


# def show_lt_vstable_rule(ns):
#    conf = config.open_config(ns.conf_path)
#    lv = logging.DEBUG if ns.debug else logging.INFO
#    config.set_common_logging(conf, logger = _logger, lv = lv)
#    from . import log_db
#    from . import lt_tool
#
#    restr = ns.word
#    update = ns.update
#    ld = log_db.LogData(conf)
#    lt_tool.search_stable_vrule(ld, restr, update)

def lttool_merge(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db
    from . import lt_tool

    ltid1 = ns.ltid1
    ltid2 = ns.ltid2
    ld = log_db.LogData(conf)
    lt_tool.merge_ltid(ld, ltid1, ltid2)


def lttool_separate(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db
    from . import lt_tool

    ltid = ns.ltid
    vid = ns.vid
    word = ns.word
    ld = log_db.LogData(conf)
    lt_tool.separate_ltid(ld, ltid, vid, word)


def lttool_split(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db
    from . import lt_tool

    ltid = ns.ltid
    vid = ns.vid
    ld = log_db.LogData(conf)
    lt_tool.split_ltid(ld, ltid, vid)


def lttool_fix(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db
    from . import lt_tool

    ltid = ns.ltid
    l_vid = ns.vids
    ld = log_db.LogData(conf)
    lt_tool.fix_ltid(ld, ltid, l_vid)


def lttool_free(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db
    from . import lt_tool

    ltid = ns.ltid
    l_wid = ns.wids
    ld = log_db.LogData(conf)
    lt_tool.free_ltid(ld, ltid, l_wid)


def lttool_fix_search(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db
    from . import lt_tool

    restr = ns.rule
    dry = ns.dry
    ld = log_db.LogData(conf)
    lt_tool.search_stable_vrule(ld, restr, dry)


def lttool_free_search(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db
    from . import lt_tool

    restr = ns.rule
    dry = ns.dry
    ld = log_db.LogData(conf)
    lt_tool.search_desc_free(ld, restr, dry)


def show_ltg_label(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import lt_label

    lt_label.list_ltlabel(conf)


def show_log_label(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import lt_label

    lt_label.count_ltlabel(conf)


def show_host(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db

    log_db.show_all_host(conf)


def show_log(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db
    lidflag = ns.lid

    d = parse_condition(ns.conditions)
    ld = log_db.LogData(conf)
    for e in ld.iter_lines(**d):
        if lidflag:
            print("{0} {1}".format(e.lid, e.restore_line()))
        else:
            print(e.restore_line())


def parse_condition(conditions):
    """
    Args:
        conditions (list)
    """
    d = {}
    for arg in conditions:
        if not "=" in arg:
            raise SyntaxError
        key = arg.partition("=")[0]
        if key == "ltid":
            d["ltid"] = int(arg.partition("=")[-1])
        elif key == "gid":
            d["ltgid"] = int(arg.partition("=")[-1])
        elif key == "top_date":
            date_string = arg.partition("=")[-1]
            d["top_dt"] = datetime.datetime.strptime(date_string, "%Y-%m-%d")
        elif key == "end_date":
            date_string = arg.partition("=")[-1]
            d["end_dt"] = datetime.datetime.strptime(date_string, "%Y-%m-%d")
        elif key == "date":
            date_string = arg.partition("=")[-1]
            d["top_dt"] = datetime.datetime.strptime(date_string, "%Y-%m-%d")
            d["end_dt"] = d["top_dt"] + datetime.timedelta(days=1)
        elif key == "host":
            d["host"] = arg.partition("=")[-1]
        elif key == "area":
            d["area"] = arg.partition("=")[-1]
    return d


def conf_defaults(ns):
    config.show_default_config()


def conf_diff(ns):
    files = ns.files[:]
    if ns.configset:
        files += config.read_config_group(ns.configset)
    config.show_config_diff(files)


def conf_minimum(ns):
    conf = config.open_config(ns.conf_path, nodefault=True, noimport=True)
    conf = config.minimize(conf)
    config.write(ns.conf_path, conf)
    print("rewrite {0}".format(ns.conf_path))


def conf_set_edit(ns):
    l_conf_name = config.read_config_group(ns.configset)
    key = ns.key
    rulestr = ns.rule

    if "(" in rulestr:
        temp = rulestr.rstrip(")").split("(")
        if not len(temp) == 2:
            raise ValueError("bad format for value specification")
        rulename, argstr = temp
        args = argstr.split(",")

        if rulename == "list":
            assert len(args) == len(l_conf_name)
            d = {key: args}
        elif rulename == "range":
            assert len(args) == 2
            start, step = [int(v) for v in args]
            l_val = [start + i * step for i in range(len(l_conf_name))]
            d = {key: l_val}
        elif rulename == "power":
            assert len(args) == 2
            start, step = [int(v) for v in args]
            l_val = [start * (i ** step) for i in range(len(l_conf_name))]
            d = {key: l_val}
        elif rulename == "withconf":
            assert len(args) == 1
            l_val = [args[0] + name for i, name in enumerate(l_conf_name)]
            d = {key: l_val}
        elif rulename == "namerange":
            assert len(args) == 1
            l_val = [args[0] + str(i) for i in range(len(l_conf_name))]
            d = {key: l_val}
        else:
            raise NotImplementedError("invalid rule name")
    else:
        l_val = [rulestr] * len(l_conf_name)
        d = {key: l_val}

    config.config_group_edit(l_conf_name, d)


def conf_shadow(ns):
    cond = {}
    incr = []
    for rule in ns.rules:
        if "=" in rule:
            key, val = rule.split("=")
            cond[key] = val
        else:
            incr.append(rule)
    l_conf_name = config.config_shadow(n=ns.number, cond=cond, incr=incr,
                                       fn=ns.conf_path, output=ns.output,
                                       ignore_overwrite=ns.force)

    if ns.configset is not None:
        config.dump_config_group(ns.configset, l_conf_name)
        print(ns.configset)


# common argument settings
OPT_DEBUG = [["--debug"],
             {"dest": "debug", "action": "store_true",
              "help": "set logging level to debug (default: info)"}]
OPT_CONFIG = [["-c", "--config"],
              {"dest": "conf_path", "metavar": "CONFIG", "action": "store",
               # "default": config.DEFAULT_CONFIG,
               "default": None,
               "help": "configuration file path for amulog"}]
OPT_CONFIG_SET = [["-s", "--configset"],
                  {"dest": "configset", "metavar": "CONFIG_SET",
                   "default": None,
                   "help": "use config group definition file"}]
OPT_RECUR = [["-r", "--recur"],
             {"dest": "recur", "action": "store_true",
              "help": "recursively search files to process"}]
OPT_DRY = [["-d", "--dry"],
             {"dest": "dry", "action": "store_true",
              "help": "do not store data into db"}]
OPT_TERM = [["-t", "--term"],
            {"dest": "dt_range",
             "metavar": "DATE1 DATE2", "nargs": 2,
             "help": ("datetime range, start and end in %Y-%M-%d style."
                      "(optional; defaultly use all data in database)")}]
ARG_FILE = [["file"],
            {"metavar": "PATH", "action": "store",
             "help": "filepath to output"}]
ARG_FILES = [["files"],
             {"metavar": "PATH", "nargs": "+",
              "help": "files or directories as input"}]
ARG_FILES_OPT = [["files"],
                 {"metavar": "PATH", "nargs": "*",
                  "help": ("files or directories as input "
                           "(optional; defaultly read from config")}]
ARG_DBSEARCH = [["conditions"],
                {"metavar": "CONDITION", "nargs": "+",
                 "help": ("Conditions to search log messages. "
                          "Example: MODE gid=24 date=2012-10-10 ..., "
                          "Keys: ltid, gid, date, top_date, end_date, "
                          "host, area")}]

# argument settings for each modes
# description, List[args, kwargs], func
# defined after functions because these settings use functions
DICT_ARGSET = {
    "data-from-db": ["Generate log data from DB.",
                     [OPT_CONFIG, OPT_DEBUG,
                      [["-d", "--dirname"],
                       {"dest": "dirname", "metavar": "DIRNAME",
                        "action": "store",
                        "help": "directory name to output"}],
                      [["-i", "--incr"],
                       {"dest": "incr", "action": "store_true",
                        "help": "output incrementally, use with small memory"}],
                      [["--reset"],
                       {"dest": "reset", "action": "store_true",
                        "help": "reset log file directory before processing"}],
                      ],
                     data_from_db],
    "data-from-data": ["Re-arrange log file, splitting messages by date.",
                       [OPT_CONFIG, OPT_DEBUG, OPT_RECUR, ARG_FILES_OPT,
                        [["-d", "--dirname"],
                         {"dest": "dirname", "metavar": "DIRNAME",
                          "action": "store",
                          "help": "directory name to output"}],
                        [["-i", "--incr"],
                         {"dest": "incr", "action": "store_true",
                          "help": "output incrementally, use with small memory"}],
                        [["--reset"],
                         {"dest": "reset", "action": "store_true",
                          "help": "reset log file directory before processing"}],
                        ],
                       data_from_data],
    "db-make": [("Initialize database and add log data. "
                 "This fuction works incrementaly."),
                [OPT_CONFIG, OPT_DEBUG, OPT_RECUR, OPT_DRY,
                 ARG_FILES_OPT],
                db_make],
    "db-add": ["Add log data to existing database.",
               [OPT_CONFIG, OPT_DEBUG, OPT_RECUR, ARG_FILES],
               db_add],
    "db-anonymize": [("Remove variables in log messages. "
                      "(Not anonymize hostnames; to be added)"),
                     [OPT_CONFIG, OPT_DEBUG],
                     db_anonymize],
    "db-reload-area": ["Reload area definition file from config.",
                       [OPT_CONFIG, OPT_DEBUG],
                       reload_area],
    "show-db-info": ["Show abstruction of database status.",
                     [OPT_CONFIG, OPT_DEBUG, OPT_TERM],
                     show_db_info],
    "show-lt": ["Show all log templates in database.",
                [OPT_CONFIG, OPT_DEBUG,
                 [["-s", "--simple"],
                  {"dest": "simple", "action": "store_true",
                   "help": "only show log templates"}]],
                show_lt],
    "show-ltg": ["Show all log template groups and their members in database.",
                 [OPT_CONFIG, OPT_DEBUG,
                  [["-g", "--group"],
                   {"dest": "group", "action": "store", "default": None,
                    "help": "show members of given labeling group"}]],
                 show_ltg],
    "show-lt-import": ["Output log template definitions in lt_import format.",
                       [OPT_CONFIG, OPT_DEBUG,
                        [["-e", "--external"],
                         {"dest": "external", "action": "store_true",
                          "help": "output in external format (for RE)"}]],
                       show_lt_import],
#    "show-lt-import-exception": [("Output log messages in a file "
#                                  "that is not defined "
#                                  "in lt_import definitions."),
#                                 [OPT_CONFIG, OPT_DEBUG, OPT_RECUR, ARG_FILES,
#                                  [["-f", "--format"],
#                                   {"dest": "format", "action": "store",
#                                    "default": "log",
#                                    "help": ("message format to load: "
#                                             "log: with header,"
#                                             "message: without header")}]],
#                                 show_lt_import_exception],
    "show-host": ["Show all hostnames in database.",
                  [OPT_CONFIG, OPT_DEBUG],
                  show_host],
    "show-ltg-label": ["Show labels for log template groups",
                       [OPT_CONFIG, OPT_DEBUG],
                       show_ltg_label],
    "show-log-label": ["Show label distributions",
                       [OPT_CONFIG, OPT_DEBUG],
                       show_log_label],
    "show-log": ["Show log messages that satisfy given conditions in args.",
                 [OPT_CONFIG, OPT_DEBUG,
                  [["--lid"],
                   {"dest": "lid", "action": "store_true",
                    "help": "show lid"}],
                  ARG_DBSEARCH],
                 show_log],
    "lttool-countall": ["Show words and their counts in all messages",
                        [OPT_CONFIG, OPT_DEBUG],
                        show_lt_words],
    "lttool-countd": ["Show description words and their counts",
                      [OPT_CONFIG, OPT_DEBUG],
                      show_lt_descriptions],
    "lttool-countv": ["Show variable words and their counts",
                      [OPT_CONFIG, OPT_DEBUG,
                       [["-d", "--digit"],
                        {"dest": "repld", "action": "store_true",
                         "help": "replace digit to \\d"}]],
                      show_lt_variables],
    "lttool-breakdown": ["Show variable candidates in a log template.",
                         [OPT_CONFIG, OPT_DEBUG,
                          [["-l", "--ltid"],
                           {"dest": "ltid", "metavar": "LTID",
                            "action": "store", "type": int,
                            "help": "log template identifier to investigate"}
                           ],
                          [["-n", "--number"],
                           {"dest": "lines", "metavar": "LINES",
                            "action": "store", "type": int, "default": 5,
                            "help": "number of variable candidates to show"}
                           ]],
                         show_lt_breakdown],
    "lttool-vstable": ["Show stable variables in the template.",
                       [OPT_CONFIG, OPT_DEBUG,
                        [["-n", "--number"],
                         {"dest": "number", "metavar": "NUMBER",
                          "action": "store", "type": int, "default": 1,
                          "help": "thureshold number to be stable"}]],
                       show_lt_vstable],
    "lttool-merge": ["Merge 2 templates and generate a new template.",
                     [OPT_CONFIG, OPT_DEBUG,
                      [["ltid1"],
                       {"metavar": "LTID1", "action": "store", "type": int,
                        "help": "first log template to merge"}],
                      [["ltid2"],
                       {"metavar": "LTID2", "action": "store", "type": int,
                        "help": "second log template to merge"}], ],
                     lttool_merge],
    "lttool-separate": ["Separate messages satisfying the given condition "
                        "and make it a new log template.",
                        [OPT_CONFIG, OPT_DEBUG,
                         [["ltid"],
                          {"metavar": "LTID", "action": "store", "type": int,
                           "help": "log template indentifier"}],
                         [["vid"],
                          {"metavar": "VARIABLE-ID", "action": "store",
                           "type": int,
                           "help": "variable identifier to fix"}],
                         [["word"],
                          {"metavar": "WORD", "action": "store",
                           "help": "a word to fix in the location of vid"}]],
                        lttool_separate],
    "lttool-split": ["Fix variables to split a template into tempaltes. "
                     "Use carefully because this function may cause "
                     "to generate enormous failure templates.",
                     [OPT_CONFIG, OPT_DEBUG,
                      [["ltid"],
                       {"metavar": "LTID", "action": "store", "type": int,
                        "help": "log template indentifier"}],
                      [["vid"],
                       {"metavar": "VARIABLE-ID", "action": "store",
                        "help": "variable identifier to fix"}]],
                     lttool_split],
    "lttool-fix": ["Fix specified stable variable and modify template. "
                   "If not stable, use lttool-split or lttool-separate.",
                   [OPT_CONFIG, OPT_DEBUG,
                    [["ltid"],
                     {"metavar": "LTID", "action": "store", "type": int,
                      "help": "log template to fix"}],
                    [["vids"],
                     {"metavar": "VARIABLE-IDs", "nargs": "+", "type": int,
                      "help": "variable identifiers to fix"}]],
                   lttool_fix],
    "lttool-free": ["Make a description word as a variable"
                    "and modify the template.",
                    [OPT_CONFIG, OPT_DEBUG,
                     [["ltid"],
                      {"metavar": "LTID", "action": "store", "type": int,
                       "help": "log template indentifier"}],
                     [["wids"],
                      {"metavar": "WORD-IDs", "nargs": "+", "type": int,
                       "help": "word locations to fix"}]],
                    lttool_free],
    "lttool-fix-search": ["Search templates with variables of given rule "
                          "and modify templates by fixing them.",
                          [OPT_CONFIG, OPT_DEBUG,
                           [["-d", "--dry-run"],
                            {"dest": "dry",
                             "action": "store_true", "default": False,
                             "help": "do not update templates"}],
                           [["rule"],
                            {"metavar": "RULE", "action": "store",
                             "help": "word / regular expression"}]],
                          lttool_fix_search],
    "lttool-free-search": ["Search templates with words of given rule ",
                           "and modify templates by making them variable.",
                           [OPT_CONFIG, OPT_DEBUG,
                            [["-u", "--update"],
                             {"dest": "update",
                              "action": "store_true", "default": False,
                              "help": "update templates automatically"}],
                            [["rule"],
                             {"metavar": "RULE", "action": "store",
                              "help": "word / regular expression"}]],
                           lttool_free_search],
    #    "make-lt-mp": [("Generate log templates with CRF "
    #                    "in multiprocessing."),
    #                   [OPT_CONFIG, OPT_DEBUG, OPT_RECUR, ARG_FILES_OPT,
    #                    [["-p", "--pal"],
    #                     {"dest": "pal", "action": "store",
    #                      "type": int, "default": 1,
    #                      "help": "number of processes"}],
    #                    [["-i", "--import"],
    #                     {"dest": "check_import", "action": "store_true",
    #                      "help": ("ignore messages corresponding to "
    #                               "imported log template definition "
    #                               "(i.e., process only unknown messages)")}],],
    #                   make_lt_mp],
    #    "measure-crf": ["Measure accuracy of CRF-based log template estimation.",
    #                    [OPT_DEBUG,
    #                     [["-f", "--failure"],
    #                      {"dest": "failure", "action": "store",
    #                       "help": "output failure report"}],
    #                     [["-c", "--config"],
    #                      {"dest": "conf_path", "metavar": "CONFIG",
    #                       "action": "store", "default": None,
    #                       "help": "extended config file for measure-lt"}],],
    #                    measure_crf],
    #    "measure-crf-multi": ["Multiprocessing of measure-crf.",
    #                          [OPT_DEBUG, OPT_CONFIG_SET,
    #                           [["-p", "--pal"],
    #                            {"dest": "pal", "action": "store",
    #                             "type": int, "default": 1,
    #                             "help": "number of processes"}],
    #                           [["-d", "--diff"],
    #                            {"dest": "diff", "action": "append",
    #                             "default": [],
    #                             "help": ("check configs that given option values "
    #                                      "are all different. This option can "
    #                                      "be specified multiple times. "
    #                                      "Example: -d general.import -d ...")}],
    #                           [["confs"],
    #                            {"metavar": "CONFIG", "nargs": "*",
    #                             "help": "configuration files"}]],
    #                          measure_crf_multi],
    "conf-defaults": ["Show default configurations.",
                      [],
                      conf_defaults],
    "conf-diff": ["Show differences of 2 configuration files.",
                  [OPT_CONFIG_SET,
                   [["files"],
                    {"metavar": "FILENAME", "nargs": "*",
                     "help": "configuration file"}]],
                  conf_diff],
    "conf-minimum": ["Remove default options and comments.",
                     [[["-o", "--overwrite"],
                       {"dest": "overwrite", "action": "store_true",
                        "help": "overwrite file instead of stdout dumping"}],
                      [["conf_path"],
                       {"metavar": "PATH",
                        "help": "config filepath to load"}]],
                     conf_minimum],
    "conf-group-edit": ["Edit configuration files in a config group.",
                        [[["configset"],
                          {"metavar": "CONFIG_SET",
                           "help": "config group definition file to use"}],
                         [["key"],
                          {"metavar": "KEY",
                           "help": "\"SECTION.OPTION\" to edit"}],
                         [["rule"],
                          {"metavar": "RULE",
                           "help": ("Value specification rule "
                                    "defined in function-like format. "
                                    "Example: \"List(1,10,100,1000)\" "
                                    "Available format: "
                                    "list(values for each config), "
                                    "range(START,STEP), "
                                    "power(START,STEP), "
                                    "withconf(NAME),"
                                    "namerange(NAME)."
                                    "Note that 1 rule for 1 execution.")}]],
                        conf_set_edit],
    "conf-shadow": ["Copy configuration files.",
                    [OPT_CONFIG,
                     [["-f", "--force"],
                      {"dest": "force", "action": "store_true",
                       "help": "Ignore overwrite of output file"}],
                     [["-n", "--number"],
                      {"dest": "number", "metavar": "INT",
                       "action": "store", "type": int, "default": 1,
                       "help": "number of files to generate"}],
                     [["-o", "--output"],
                      {"dest": "output", "metavar": "FILENAME",
                       "action": "store", "type": str, "default": None,
                       "help": "basic output filename"}],
                     [["-s", "--configset"],
                      {"dest": "configset", "metavar": "CONFIG_SET",
                       "default": None,
                       "help": ("define config group "
                                "and dump it in given filename")}],
                     [["rules"],
                      {"metavar": "RULES", "nargs": "*",
                       "help": ("Rules to replace options. You can indicate "
                                "option, or option and its value with =. "
                                "You can use both of them together. "
                                "For example: \"general.import=hoge.conf "
                                "general.logging\"")}]],
                    conf_shadow],
}

USAGE_COMMANDS = "\n".join(["  {0}: {1}".format(key, val[0])
                            for key, val in sorted(DICT_ARGSET.items())])
USAGE = ("usage: {0} MODE [options and arguments] ...\n\n"
         "mode:\n".format(sys.argv[0])) + USAGE_COMMANDS + \
        "\n\nsee \"{0} MODE -h\" to refer detailed usage".format(sys.argv[0]) + \
        "\nalso see sub-liblary {0}".format(" ".join(["logdag.{0}".format(n)
                                                      for n in SUBLIB]))


def _main():
    if len(sys.argv) < 2:
        sys.exit(USAGE)
    mode = sys.argv[1]
    if mode in ("-h", "--help"):
        sys.exit(USAGE)
    commandline = sys.argv[2:]

    desc, l_argset, func = DICT_ARGSET[mode]
    ap = argparse.ArgumentParser(prog=" ".join(sys.argv[0:2]),
                                 description=desc)
    for args, kwargs in l_argset:
        ap.add_argument(*args, **kwargs)
    ns = ap.parse_args(commandline)
    func(ns)


if __name__ == "__main__":
    _main()



#!/usr/bin/env python
# coding: utf-8

"""
Construct and manage database of log messages, templates,
and grouping definitions.
"""

import sys
import os
import datetime
import logging
from collections import defaultdict
import log2seq

from . import common
from . import config
from . import strutil
from . import db_common
from . import lt_common
from . import host_alias

_logger = logging.getLogger(__package__)
ONLINE_COMMIT_INTERVAL = 1000


class LogMessage():
    """An annotated log message.
    
    An instance have a set of information about 1 log message line,
    including timestamp, hostname, log template, and message contents.

    Attributes:
        lid (int): A message identifier in DB.
        lt (lt_common.LogTemplate): Log template of this message.
        dt (datetime.datetime): A timestamp for this message.
        host (str): A hostname that output this message.
        l_w (List(str)): A sequence of words in this message.

    """

    def __init__(self, lid, lt, dt, host, l_w):
        """
        Args:
            lid (int): A message identifier in DB.
            lt (lt_common.LogTemplate): Log template of this message.
            dt (datetime.datetime): A timestamp for this message.
            host (str): A hostname that output this message.
            l_w (List(str)): A sequence of words in this message.

        """
        self.lid = lid
        self.lt = lt
        self.dt = dt
        self.host = host
        self.l_w = l_w

    def __str__(self):
        """str: Show attributes in 1 string."""
        return " ".join((str(self.dt), self.host, str(self.lt.ltid),
                         str(self.l_w)))

    def get(self, name):
        """Return value for given attribute."""
        if name == "lid":
            return self.lid
        elif name == "ltid":
            return self.lt.ltid
        elif name == "ltgid":
            return self.lt.ltgid
        elif name == "host":
            return self.host
        else:
            raise NotImplementedError

    def var(self):
        """str: Get sequence of all variable words in this message.
        Variable words mean what is presented with mask
        (defaults **) in log template.
        """
        return self.lt.var(self.l_w)

    def restore_message(self):
        """str: Get pseudo original log message
        including headers (timestamp and hostname).
        """
        return self.lt.restore_message(self.l_w)

    def restore_line(self):
        """str: Get original log message contents
        except headers (timestamp and hostname).
        """
        return " ".join((str(self.dt), str(self.host),
                         self.restore_message()))

    # def restore_line_lid(self):
    #    """str: Get original log message contents
    #    except headers (timestamp and hostname).
    #    """
    #    return " ".join((str(self.lid), self.restore_line()))


class LogData:
    """Interface to get, add or edit log messages in DB.

    Attributes:
        conf (config.ExtendedConfigParser): A common configuration object.
        lttable (lt_common.LTTable): Log template table object.
        db (LogDB): Log database instance.
        ltm (lt_common.LTManager): Log template classifier object.
    """

    def __init__(self, conf, edit=False, reset_db=False):
        """
        Args:
            conf (config.ExtendedConfigParser): A common configuration object.
            edit (Optional[bool]): Defaults to False.
                True if database will be added or edited.
                False if database is used in readonly mode.
            reset_db (Optional[bool]): Defaults to False.
                If True, database will be reset before following process.

        """
        self.conf = conf
        self._reset_db = reset_db
        self.lttable = lt_common.LTTable()  # lt_common.LTTable
        self.db = LogDB(conf, self.lttable, edit, reset_db)  # log_db.LogDB
        self.ltm = None  # lt_common.LTManager
        from . import lt_label
        self.ll = lt_label.init_ltlabel(conf)

    def init_ltmanager(self):
        """Initialize log template classifier object.
        Call this before adding new log message in original
        plain string format (not classified with log template) to DB.
        """
        if self.ltm is None:
            self.ltm = lt_common.init_ltmanager(self.conf, self.db,
                                                self.lttable, self._reset_db)

    def get_line(self, lid):
        return self.db.get_line(lid)

    def iter_lines(self, **kargs):
        """Generate log messages in DB that satisfy conditions
        given in arguments. All arguments are defaults to None, and ignored.

        Args:
            lid (Optional[int]): A message identifier in DB.
            ltid (Optional[int]): A log template identifier.
            ltgid (Optional[int]): A log template grouping identifier.
            top_dt (Optional[datetime.datetime]): Condition for timestamp.
                Messages output after 'top_dt' will be yield.
            end_dt (Optional[datetime.datetime]): Condition for timestamp.
                Messages output before 'end_dt' will be yield.
            host (Optional[str]): A source hostname of the message.
            area (Optional[str]): An area name of source hostname.
                Some reserved area-name is available;
                \"all\" to use whole data for 1 area,
                \"host_*\" to use 1 hostname as 1 area.
                (\"each\" is only used for pc_log argument initialization)

        Yields:
            LogMessage: An annotated log message instance
                which satisfies all given conditions.
        """
        _logger.debug("iter_lines called ({0})".format(" ".join(
            ["{0}:{1}".format(k, v) for k, v in kargs.items()
             if v is not None])))
        return self.db.iter_lines(**kargs)

    def count_lines(self):
        """int: Number of all messages in DB."""
        return self.db.count_lines()

    def dt_term(self):
        """datetime.datetime, datetime.datetime:
            Period of all registered log messages in DB."""
        return self.db.dt_term()

    def whole_term(self):
        """datetime.datetime, datetime.datetime:
            Period of dates that include all log messages in DB.
        """
        # return date term that cover all log data
        top_dt, end_dt = self.db.dt_term()
        top_dt = datetime.datetime.combine(top_dt.date(), datetime.time())
        end_dt = datetime.datetime.combine(end_dt.date(), datetime.time()) + \
                 datetime.timedelta(days=1)
        return top_dt, end_dt

    def whole_host_lt(self, top_dt=None, end_dt=None, area=None):
        """List[str, str]: Sequence of all combinations of 
        hostname and ltids in DB."""
        return self.db.whole_host_lt(top_dt=top_dt, end_dt=end_dt,
                                     area=area)

    def whole_host_ltg(self, top_dt=None, end_dt=None, area=None):
        """List[str, str]: Sequence of all combinations of 
        hostname and ltgids in DB."""
        ret = set()
        for host, ltid in self.whole_host_lt(top_dt=top_dt, end_dt=end_dt,
                                             area=area):
            ret.add((host, self.ltgid_from_ltid(ltid)))
        return list(set(ret))

    def whole_host(self, top_dt=None, end_dt=None):
        """List[str]: Sequence of all source hostname in DB."""
        return self.db.whole_host(top_dt=top_dt, end_dt=end_dt)

    def count_lt(self):
        """int: Number of all log templates."""
        return self.db.count_lt()

    def count_ltg(self):
        """int: Number of all log template groups."""
        return self.db.count_ltg()

    def iter_lt(self):
        """Yields lt_common.LogTemplate: All log template instance in DB."""
        for ltline in self.lttable:
            yield ltline

    def lt(self, ltid):
        """Get log template instance of given log template identifier.

        Args:
            ltid (int): A log template identifier.

        Returns:
            lt_common.LogTemplate: A log template instance.

        """
        return self.lttable[ltid]

    def iter_gid(self, gid_name):
        if gid_name == "ltid":
            return [lt.ltid for lt in self.iter_lt()]
        elif gid_name == "ltgid":
            return self.iter_ltgid()

    def iter_ltgid(self):
        """Yields int: Get all identifiers of log template groups."""
        return self.db.iter_ltgid()

    def ltgid_from_ltid(self, ltid):
        lt = self.lttable[ltid]
        return lt.ltgid

    def ltg_members(self, ltgid):
        """Get all log templates in given log template group.
        
        Args:
            ltgid (int): A log template group identifier.

        Returns:
            List[lt_common.LogTemplates]: Sequence of all log template
                instances that belongs to given log template group.
        
        """
        return [self.lttable[ltid] for ltid in self.db.get_ltg_members(ltgid)]

    def host_area(self, host):
        """Get area names that given host belongs to.
        
        Args:
            host (str): A hostname.

        Returns:
            List[str]: A sequence of area names that given host belongs to.

        """
        return self.db.host_area(host)

    @staticmethod
    def _str_ltline(ltline):
        return " ".join((str(ltline.ltid), "({0})".format(ltline.ltgid),
                         str(ltline), "({0})".format(ltline.cnt)))

    def show_template_table(self):
        """For debugging"""
        self.init_ltmanager()
        return self.ltm._table

    def dump_template_table(self):
        """For debugging"""
        self.init_ltmanager()
        return "\n".join([" ".join(tpl) for tpl in self.ltm._table])

    def show_all_lt(self):
        """Show all log templates. Log template identifier
        and its template message will be output.

        Returns:
            str: Output message buffer.
        """
        buf = []
        for ltline in self.lttable:
            buf.append(self._str_ltline(ltline))
        return "\n".join(buf)

    def show_all_ltgroup(self):
        """Show all log template groups. Log template grouping identifier
        and its template candidates will be output.

        Returns:
            str: Output message buffer.
        """
        if self.db.count_ltg() == 0:
            self.show_all_lt()
        else:
            buf = []
            for ltgid in self.db.iter_ltgid():
                buf.append(self.show_ltgroup(ltgid))
            return "\n".join(buf)

    def show_ltgroup_cond(self, **kwargs):
        buf = []
        for gid in self.db.iter_ltgid():
            l_lt = self.ltg_members(gid)
            l_ltid = self.db.get_ltg_members(gid)
            label = self.ll.get_ltg_label(gid, l_lt)
            group = self.ll.get_ltg_group(gid, l_lt)
            if "group" in kwargs and not kwargs["group"] == group:
                continue
            if "label" in kwargs and not kwargs["label"] == label:
                continue
            buf.append(self.show_ltgroup(gid))
        return "\n".join(buf)

    def show_ltgroup(self, gid):
        """Show log template groups.

        Returns:
            str: Output message buffer.
        """
        buf = []
        l_lt = self.ltg_members(gid)
        l_ltid = self.db.get_ltg_members(gid)
        label = self.ll.get_ltg_label(gid, l_lt)
        group = self.ll.get_ltg_group(gid, l_lt)
        length = len(l_ltid)
        cnt = 0
        for ltid in l_ltid:
            ltline = self.lttable[ltid]
            cnt += ltline.cnt
            buf.append(self._str_ltline(ltline))

        buf = ["[ltgroup {0} ({1}, {2})] # {3}({4})".format(gid, length,
                                                            cnt, group,
                                                            label)] + buf
        return "\n".join(buf)

    def add_line(self, ltid, pline):
        """Add a log message to DB.

        Args:
            ltid (int): A log template identifier.
            pline (dict): A parsed log message with log2seq.

        Returns:
            LogMessage: An annotated log message instance.
        """
        kwargs = {"ltid": ltid,
                  "dt": pline["timestamp"],
                  "host": pline["host"],
                  "l_w": pline["words"]}
        if "lid" in pline:
            kwargs["lid"] = pline["lid"]
        new_lid = self.db.add_line(**kwargs)
        return LogMessage(new_lid, self.lttable[ltid],
                          pline["timestamp"], pline["host"], pline["words"])

    def update_area(self):
        self.db._remove_area()
        self.db._init_area()

    def commit_db(self):
        """Commit requested changes in LogDB.
        """
        self.db.commit()
        if self.ltm is not None:
            self.ltm.dump()


class LogDB:
    """Interface of DB transaction for log data.
    
    Note:
        It is not recommended to use this class directly.
        Instead, use LogData.
    """

    def __init__(self, conf, lttable, edit, reset_db):
        self.lttable = lttable
        self._line_cnt = 0
        self.areafn = conf.get("database", "area_filename")
        self._splitter = conf.get("database", "split_symbol")

        db_type = conf.get("database", "database")
        if db_type == "sqlite3":
            dbpath = conf.get("database", "sqlite3_filename")
            if dbpath is None:
                # for compatibility
                dbpath = conf.get("database", "db_filename")
            self.db = db_common.sqlite3(dbpath)
        elif db_type == "mysql":
            host = conf.get("database", "mysql_host")
            dbname = conf.get("database", "mysql_dbname")
            user = conf.get("database", "mysql_user")
            passwd = conf.get("database", "mysql_passwd")
            self.db = db_common.mysql(host, dbname, user, passwd)
        else:
            raise ValueError("invalid database type ({0})".format(
                db_type))

        if edit:
            if self.db.db_exists():
                if reset_db:
                    self.db.reset()
                    self._init_tables()
                    self._init_area()
                else:
                    self._line_cnt = self.count_lines()
                    self._init_lttable()
            else:
                if reset_db:
                    _logger.warning(
                        "Requested to reset DB, but database not found")
                self._init_tables()
                self._init_area()
        else:
            if self.db.db_exists():
                self._line_cnt = self.count_lines()
                self._init_lttable()
            else:
                raise IOError("database not found")

    def _init_tables(self):
        table_name = "log"
        l_key = [db_common.tablekey("lid", "integer",
                                    # ("primary_key", "auto_increment", "not_null")),
                                    ("primary_key", "not_null")),
                 db_common.tablekey("ltid", "integer"),
                 db_common.tablekey("dt", "datetime"),
                 db_common.tablekey("host", "text"),
                 db_common.tablekey("words", "text")]
        sql = self.db.create_table_sql(table_name, l_key)
        self.db.execute(sql)

        table_name = "lt"
        l_key = [db_common.tablekey("ltid", "integer", ("primary_key",)),
                 db_common.tablekey("ltw", "text"),
                 db_common.tablekey("lts", "text"),
                 db_common.tablekey("count", "integer")]
        sql = self.db.create_table_sql(table_name, l_key)
        self.db.execute(sql)

        table_name = "ltg"
        l_key = [db_common.tablekey("ltid", "integer", ("primary_key",)),
                 db_common.tablekey("ltgid", "integer")]
        sql = self.db.create_table_sql(table_name, l_key)
        self.db.execute(sql)

        table_name = "area"
        l_key = [db_common.tablekey("defid", "integer",
                                    ("primary_key", "auto_increment", "not_null")),
                 db_common.tablekey("host", "text"),
                 db_common.tablekey("area", "text")]
        sql = self.db.create_table_sql(table_name, l_key)
        self.db.execute(sql)

        self._init_index()

    def _init_index(self):
        l_table_name = self.db.get_table_names()

        table_name = "log"
        index_name = "log_index"
        l_key = [db_common.tablekey("lid", "integer"),
                 db_common.tablekey("ltid", "integer"),
                 db_common.tablekey("dt", "datetime"),
                 db_common.tablekey("host", "text", (100,))]
        if not index_name in l_table_name:
            sql = self.db.create_index_sql(table_name, index_name, l_key)
            self.db.execute(sql)

        table_name = "ltg"
        index_name = "ltg_index"
        l_key = [db_common.tablekey("ltgid", "integer")]
        if not index_name in l_table_name:
            sql = self.db.create_index_sql(table_name, index_name, l_key)
            self.db.execute(sql)

        table_name = "area"
        index_name = "area_index"
        l_key = [db_common.tablekey("area", "text", (100,))]
        if not index_name in l_table_name:
            sql = self.db.create_index_sql(table_name, index_name, l_key)
            self.db.execute(sql)

    def commit(self):
        self.db.commit()

    def add_line(self, ltid, dt, host, l_w, lid=None):
        table_name = "log"
        d_val = {
            "ltid": ltid,
            "dt": self.db.strftime(dt),
            "host": host,
            "words": self._splitter.join(l_w),
        }

        self._line_cnt += 1
        if lid is None:
            d_val["lid"] = self._line_cnt
        else:
            d_val["lid"] = lid

        l_ss = [db_common.setstate(k, k) for k in d_val.keys()]
        sql = self.db.insert_sql(table_name, l_ss)
        self.db.execute(sql, d_val)

        return d_val["lid"]

    def iter_lines(self, lid=None, ltid=None, ltgid=None, top_dt=None,
                   end_dt=None, host=None, area=None):
        d_cond = {}
        if lid is not None: d_cond["lid"] = lid
        if ltid is not None: d_cond["ltid"] = ltid
        if ltgid is not None: d_cond["ltgid"] = ltgid
        if top_dt is not None: d_cond["top_dt"] = top_dt
        if end_dt is not None: d_cond["end_dt"] = end_dt
        if area is None or area == "all":
            pass
        elif area[:5] == "host_":
            d_cond["host"] = area[5:]
        else:
            d_cond["area"] = area
        if host is not None: d_cond["host"] = host

        if len(d_cond) == 0:
            raise ValueError("More than 1 argument should NOT be None")

        for row in self._select_log(d_cond):
            lid = int(row[0])
            ltid = int(row[1])
            dt = self.db.datetime(row[2])
            host = row[3]
            if row[4] == "":
                l_w = []
            else:
                l_w = strutil.split_igesc(row[4], self._splitter)
            yield LogMessage(lid, self.lttable[ltid], dt, host, l_w)

    def iter_words(self, lid=None, ltid=None, ltgid=None, top_dt=None,
                   end_dt=None, host=None, area=None):
        d_cond = {}
        if lid is not None: d_cond["lid"] = lid
        if ltid is not None: d_cond["ltid"] = ltid
        if ltgid is not None: d_cond["ltgid"] = ltgid
        if top_dt is not None: d_cond["top_dt"] = top_dt
        if end_dt is not None: d_cond["end_dt"] = end_dt
        if host is not None: d_cond["host"] = host
        if area is not None: d_cond["area"] = area
        if len(d_cond) == 0:
            raise ValueError("More than 1 argument should NOT be None")
        for row in self._select_log(d_cond):
            if row[4] == "":
                yield []
            else:
                yield strutil.split_igesc(row[4], self._splitter)

    def _select_log(self, d_cond):
        if len(d_cond) == 0:
            raise ValueError("called select with empty condition")
        args = d_cond.copy()

        table_name = "log"
        l_key = ["lid", "ltid", "dt", "host", "words"]
        l_cond = []
        for c in d_cond.keys():
            if c == "ltgid":
                sql = self.db.select_sql("ltg", ["ltid"],
                                         [db_common.cond(c, "=", c)])
                l_cond.append(db_common.cond("ltid", "in", sql, False))
            elif c == "area":
                sql = self.db.select_sql("area", ["host"],
                                         [db_common.cond(c, "=", c)])
                l_cond.append(db_common.cond("host", "in", sql, False))
            elif c == "top_dt":
                l_cond.append(db_common.cond("dt", ">=", c))
                args[c] = self.db.strftime(d_cond[c])
            elif c == "end_dt":
                l_cond.append(db_common.cond("dt", "<", c))
                args[c] = self.db.strftime(d_cond[c])
            else:
                l_cond.append(db_common.cond(c, "=", c))
        sql = self.db.select_sql(table_name, l_key, l_cond)
        return self.db.execute(sql, args)

    def get_line(self, lid):
        table_name = "log"
        l_key = ["lid", "ltid", "dt", "host", "words"]
        l_cond = [db_common.cond("lid", "=", "lid")]
        sql = self.db.select_sql(table_name, l_key, l_cond)
        args = {"lid": lid}

        ret = []
        for row in self.db.execute(sql, args):
            lid = int(row[0])
            ltid = int(row[1])
            dt = self.db.datetime(row[2])
            host = row[3]
            if row[4] == "":
                l_w = []
            else:
                l_w = strutil.split_igesc(row[4], self._splitter)
            lm = LogMessage(lid, self.lttable[ltid], dt, host, l_w)
            ret.append(lm)

        if len(ret) == 0:
            return None
        elif len(ret) == 1:
            return ret[0]
        else:
            raise ValueError("Duplicated messages for lid {0}".format(lid))

    def update_log(self, d_cond, d_update):
        if len(d_cond) == 0:
            _logger.warning("called update with empty condition")
            # raise ValueError("called update with empty condition")
        args = d_cond.copy()

        table_name = "log"
        l_ss = []
        for k, v in d_update.items():
            # assert k in ("ltid", "top_dt", "end_dt", "host")
            keyname = "update_" + k
            l_ss.append(db_common.setstate(k, keyname))
            args[keyname] = v
        l_cond = []
        for c in d_cond.keys():
            if c == "ltgid":
                sql = self.db.select_sql("ltg", ["ltid"],
                                         [db_common.cond(c, "=", c)])
                l_cond.append(db_common.cond("ltid", "in", sql, False))
            elif c == "area":
                sql = self.db.select_sql("area", ["host"],
                                         [db_common.cond(c, "=", c)])
                l_cond.append(db_common.cond("host", "in", sql, False))
            elif c == "top_dt":
                l_cond.append(db_common.cond("dt", ">=", c))
                args[c] = self.db.strftime(d_cond[c])
            elif c == "end_dt":
                l_cond.append(db_common.cond("dt", "<", c))
                args[c] = self.db.strftime(d_cond[c])
            else:
                l_cond.append(db_common.cond(c, "=", c))
        sql = self.db.update_sql(table_name, l_ss, l_cond)
        self.db.execute(sql, args)

    def count_lines(self):
        table_name = "log"
        l_key = ["max(lid)"]
        sql = self.db.select_sql(table_name, l_key)
        cursor = self.db.execute(sql)
        tmp = cursor.fetchone()[0]
        if tmp is None:
            return 0
        else:
            return int(tmp)

    def dt_term(self):
        table_name = "log"
        l_key = ["min(dt)", "max(dt)"]
        sql = self.db.select_sql(table_name, l_key)
        cursor = self.db.execute(sql)
        top_dtstr, end_dtstr = cursor.fetchone()
        if None in (top_dtstr, end_dtstr):
            raise ValueError("No data found in DB")
        return self.db.datetime(top_dtstr), self.db.datetime(end_dtstr)

    def whole_host_lt(self, top_dt=None, end_dt=None, area=None):
        table_name = "log"
        l_key = ["host", "ltid"]
        l_cond = []
        args = {}
        if top_dt is not None:
            l_cond.append(db_common.cond("dt", ">=", "top_dt"))
            # args["top_dt"] = top_dt
            args["top_dt"] = self.db.strftime(top_dt)
        if end_dt is not None:
            l_cond.append(db_common.cond("dt", "<", "end_dt"))
            # args["end_dt"] = end_dt
            args["end_dt"] = self.db.strftime(end_dt)
        if area is None or area == "all":
            pass
        elif area[:5] == "host_":
            l_cond.append(db_common.cond("host", "=", "host"))
            args["host"] = area[5:]
        else:
            temp_sql = self.db.select_sql(
                "area", ["host"], [db_common.cond("area", "=", "area")])
            l_cond.append(db_common.cond("host", "in", temp_sql, False))
            args["area"] = area

        sql = self.db.select_sql(table_name, l_key, l_cond, opt=["distinct"])
        cursor = self.db.execute(sql, args)
        return [(row[0], row[1]) for row in cursor]

    def whole_host(self, top_dt=None, end_dt=None, area=None):
        table_name = "log"
        l_key = ["host"]
        l_cond = []
        args = {}
        if top_dt is not None:
            l_cond.append(db_common.cond("dt", ">=", "top_dt"))
            args["top_dt"] = self.db.strftime(top_dt)
            # args["top_dt"] = top_dt
        if end_dt is not None:
            l_cond.append(db_common.cond("dt", "<", "end_dt"))
            # args["end_dt"] = end_dt
            args["end_dt"] = self.db.strftime(end_dt)
        sql = self.db.select_sql(table_name, l_key, l_cond, opt=["distinct"])
        cursor = self.db.execute(sql, args)
        return [row[0] for row in cursor]

    def add_lt(self, ltline):
        table_name = "lt"
        l_ss = []
        l_ss.append(db_common.setstate("ltid", "ltid"))
        l_ss.append(db_common.setstate("ltw", "ltw"))
        l_ss.append(db_common.setstate("lts", "lts"))
        l_ss.append(db_common.setstate("count", "count"))
        if ltline.lts is None:
            lts = None
        else:
            lts = self._splitter.join(ltline.lts)
        args = {
            "ltid": ltline.ltid,
            "ltw": self._splitter.join(ltline.ltw),
            "lts": lts,
            "count": ltline.cnt,
        }
        sql = self.db.insert_sql(table_name, l_ss)
        self.db.execute(sql, args)

        self.add_ltg(ltline.ltid, ltline.ltgid)

    def add_ltg(self, ltid, ltgid):
        table_name = "ltg"
        l_ss = []
        l_ss.append(db_common.setstate("ltid", "ltid"))
        l_ss.append(db_common.setstate("ltgid", "ltgid"))
        args = {"ltid": ltid, "ltgid": ltgid}
        sql = self.db.insert_sql(table_name, l_ss)
        self.db.execute(sql, args)

    def update_lt(self, ltid, ltw, lts, count):
        table_name = "lt"
        l_ss = []
        args = {}
        if ltw is not None:
            l_ss.append(db_common.setstate("ltw", "ltw"))
            args["ltw"] = self._splitter.join(ltw)
        if lts is not None:
            l_ss.append(db_common.setstate("lts", "lts"))
            args["lts"] = self._splitter.join(lts)
        if count is not None:
            l_ss.append(db_common.setstate("count", "count"))
            args["count"] = count
        l_cond = [db_common.cond("ltid", "=", "ltid")]
        args["ltid"] = ltid

        sql = self.db.update_sql(table_name, l_ss, l_cond)
        self.db.execute(sql, args)

    def remove_lt(self, ltid):
        args = {"ltid": ltid}

        # remove from lt
        table_name = "lt"
        l_cond = [db_common.cond("ltid", "=", "ltid")]
        sql = self.db.delete_sql(table_name, l_cond)
        self.db.execute(sql, args)

        # remove from ltg
        table_name = "ltg"
        l_cond = [db_common.cond("ltid", "=", "ltid")]
        sql = self.db.delete_sql(table_name, l_cond)
        self.db.execute(sql, args)

    def _init_lttable(self):
        table_name = self.db.join_sql("left outer",
                                      "lt", "ltg", "ltid", "ltid")
        l_key = ("lt.ltid", "ltg.ltgid", "lt.ltw", "lt.lts", "lt.count")
        sql = self.db.select_sql(table_name, l_key)
        cursor = self.db.execute(sql)
        for row in cursor:
            ltid = int(row[0])
            ltgid = int(row[1])
            ltw = strutil.split_igesc(row[2], self._splitter)
            temp = row[3]
            if temp is None:
                lts = None
            else:
                lts = strutil.split_igesc(temp, self._splitter)
            count = int(row[4])
            self.lttable.restore_lt(ltid, ltgid, ltw, lts, count)

    def count_lt(self):
        table_name = "lt"
        l_key = ["count(*)"]
        sql = self.db.select_sql(table_name, l_key)
        cursor = self.db.execute(sql)
        return int(cursor.fetchone()[0])

    def count_ltg(self):
        table_name = "ltg"
        l_key = ["max(ltgid)"]
        sql = self.db.select_sql(table_name, l_key)
        cursor = self.db.execute(sql)
        return int(cursor.fetchone()[0]) + 1

    def iter_ltg_def(self):
        table_name = "ltg"
        l_key = ["ltid", "ltgid"]
        sql = self.db.select_sql(table_name, l_key)
        cursor = self.db.execute(sql)
        for row in cursor:
            ltid, ltgid = row
            yield int(ltid), int(ltgid)

    def iter_ltgid(self):
        table_name = "ltg"
        l_key = ["ltgid"]
        sql = self.db.select_sql(table_name, l_key, opt=["distinct"])
        cursor = self.db.execute(sql)
        for row in cursor:
            ltgid = row[0]
            yield int(ltgid)

    def get_ltg_members(self, ltgid):
        table_name = "ltg"
        l_key = ["ltid"]
        l_cond = [db_common.cond("ltgid", "=", "ltgid")]
        args = {"ltgid": ltgid}
        sql = self.db.select_sql(table_name, l_key, l_cond)
        cursor = self.db.execute(sql, args)
        return [int(row[0]) for row in cursor]

    def reset_ltg(self):
        sql = self.db.delete_sql("ltg")
        self.db.execute(sql)

    def _init_area(self):
        if self.areafn is None or self.areafn == "":
            return
        areadict = config.GroupDef(self.areafn)
        table_name = "area"
        l_ss = [db_common.setstate("host", "host"),
                db_common.setstate("area", "area")]
        sql = self.db.insert_sql(table_name, l_ss)
        for area, host in areadict.iter_def():
            args = {
                "host": host,
                "area": area
            }
            self.db.execute(sql, args)
        self.commit()

    def _remove_area(self):
        table_name = "area"
        sql = self.db.delete_sql(table_name)
        self.db.execute(sql)

    def host_area(self, host):
        table_name = "area"
        l_key = ["area"]
        l_cond = [db_common.cond("host", "=", "host")]
        args = {"host": host}
        sql = self.db.select_sql(table_name, l_key, l_cond)
        cursor = self.db.execute(sql, args)
        return [row[0] for row in cursor]


class RestoreOriginalData(object):

    def __init__(self, dirname, style="date", method="commit",
                 reset=False):
        self.style = style
        self.method = method
        self.dirname = dirname
        common.mkdir(dirname)
        if reset:
            _logger.info("RestoreOriginalData was requested to reset files")
            common.rm_dirchild(dirname)

        assert self.style in ("date", "")
        assert self.method in ("incremental", "commit")
        if self.method == "commit":
            self.buf = defaultdict(list)

    def add(self, lm):
        if self.style == "date":
            fn = lm.dt.strftime("%Y%m%d")
        else:
            raise NotImplementedError

        linestr = lm.restore_line()
        if self.method == "incremental":
            self.write_line(linestr, fn)
        elif self.method == "commit":
            self.buf[fn].append(linestr)

    def add_str(self, dt, linestr):
        if self.style == "date":
            fn = dt.strftime("%Y%m%d")
        else:
            raise NotImplementedError

        if self.method == "incremental":
            self.write_line(linestr, fn)
        elif self.method == "commit":
            self.buf[fn].append(linestr)

    def commit(self):
        if self.method == "incremental":
            pass
        elif self.method == "commit":
            self.write_all()

    def write_line(self, linestr, fn):
        with open("/".join((self.dirname, fn)),
                  "a", encoding='utf-8') as f:
            f.write(linestr + "\n")

    def write_all(self):
        assert self.method == "commit"
        for fn, l_buf in self.buf.items():
            with open("/".join((self.dirname, fn)),
                      "a", encoding='utf-8') as f:
                f.write("\n".join(l_buf))


class FailureLog:

    def __init__(self, fp):
        self._fp = fp

    def add(self, msg):
        with open(self._fp, 'a') as f:
            f.write(msg)


def load_log2seq(conf):
    """Amulog accepts following additional keys extracted by log2seq.
    - host: device hostname, mandatory
    - lid: Log message identifier, optional

    Args:
        conf:

    Returns:
        log2seq.LogParser

    """
    fp = conf.get("database", "parser_script")
    if len(fp.strip()) == 0:
        return log2seq.init_parser()
    else:
        rules = log2seq.load_from_script(fp)
        return log2seq.init_parser(rules)


def log2seq_weight_save(pline):
    unused_keys = ["year", "month", "day", "hour", "minute", "second", "tz"]
    for key in unused_keys:
        if key in pline:
            del pline[key]


def init_failure_log(conf):
    return FailureLog(conf["general"]["fail_output"])


def iter_files(targets):
    for fp in targets:
        if os.path.isdir(fp):
            sys.stderr.write(
                "{0} is a directory, fail to process\n".format(fp))
            sys.stderr.write(
                "Use -r if you need to search log data recursively\n")
        else:
            if not os.path.isfile(fp):
                raise IOError("File {0} not found".format(fp))
            with open(fp, 'r', encoding='utf-8') as f:
                _logger.info("log_db processing file {0}".format(fp))
                yield f


def parse_line(msg, lp):
    try:
        parsed_line = lp.process_line(strutil.add_esc(msg))
    except SyntaxError as e:
        _logger.info(str(e))
        return

    try:
        l_w = parsed_line["words"]
    except KeyError:
        _logger.debug("pass empty message {0}".format(parsed_line["message"]))
        return None
    if len(l_w) == 0:
        _logger.debug("pass empty message {0}".format(parsed_line["message"]))
        return None

    return parsed_line


def normalize_host(msg, pline, ha, fl, drop_undefhost=False):
    if "lid" in pline:
        pline["lid"] = int(pline["lid"])

    org_host = pline["host"]
    host = ha.resolve_host(org_host)
    if host is None:
        if drop_undefhost:
            if fl is not None:
                fl.add(msg)
            return None
        else:
            host = org_host
    pline["host"] = host

    return pline


def iter_plines(conf, targets):
    from amulog import host_alias
    lp = load_log2seq(conf)
    ha = host_alias.init_hostalias(conf)
    drop_undefhost = conf.getboolean("database", "undefined_host")

    for f in iter_files(targets):
        for msg in f:
            pline = parse_line(msg, lp)
            if pline is None:
                continue
            pline = normalize_host(msg, pline, ha, None, drop_undefhost)
            if pline is None:
                pass
            else:
                log2seq_weight_save(pline)
                yield pline


def process_files_online(conf, targets, reset_db, dry=False):
    """Add log messages to DB from files.

    Args:
        conf (config.ExtendedConfigParser): A common configuration object.
        targets (List[str]): A sequence of filepaths to process.
        reset_db (bool): True if DB needs to reset before adding.
        dry (Optional[bool]): for dry-run (no database-io)

    Raises:
        IOError: If a file in targets not found.
    """
    ld = LogData(conf, edit=True, reset_db=reset_db)
    ld.init_ltmanager()
    fl = init_failure_log(conf)

    for mid, pline in enumerate(iter_plines(conf, targets)):
        _logger.debug("Processing [{0}]".format(" ".join(pline["words"])))
        ltline = ld.ltm.process_line(pline)
        if ltline is None:
            fl.add(pline["message"])
        elif dry:
            pass
        else:
            _logger.debug("Template [{0}]".format(ltline))
            ld.add_line(ltline.ltid, pline)
        if mid % ONLINE_COMMIT_INTERVAL == 0:
            ld.commit_db()
    ld.commit_db()


def process_files_offline(conf, targets, dry=False):
    """Add log messages to DB from files. This function do NOT process
    messages incrementally. Use this to avoid bad-start problem of
    log template generation with clustering or training methods.

    Note:
        This function needs large memory space.

    Args:
        conf (config.ExtendedConfigParser): A common configuration object.
        targets (List[str]): A sequence of filepaths to process.
        dry (Optional[bool]): for dry-run (no database-io)

    Raises:
        IOError: If a file in targets not found.
    """
    ld = LogData(conf, edit=True, reset_db=True)
    ld.init_ltmanager()
    fl = init_failure_log(conf)

    l_line = [pline for pline in iter_plines(conf, targets)]

    for ltline, pline, in zip(ld.ltm.process_offline(l_line),
                              l_line):
        if ltline is None:
            fl.add(pline["message"])
        elif dry:
            pass
        else:
            ld.add_line(ltline.ltid, pline)

    ld.commit_db()


def info(conf):
    """Show abstruction of log messages registered in DB.

    Args:
        conf (config.ExtendedConfigParser): A common configuration object.

    """

    ld = LogData(conf)
    print("[DB status]")
    print("Registered log lines : {0}".format(ld.count_lines()))
    print("Term : {0[0]} - {0[1]}".format(ld.dt_term()))
    print("Log templates : {0}".format(ld.count_lt()))
    print("Log template groups : {0}".format(ld.count_ltg()))
    print("Hosts : {0}".format(len(ld.whole_host())))


def info_term(conf, top_dt, end_dt):
    cnt_line = 0
    s_ltid = set()
    s_gid = set()
    s_host = set()

    ld = LogData(conf)
    for line in ld.iter_lines(top_dt=top_dt, end_dt=end_dt):
        cnt_line += 1
        s_ltid.add(line.lt.ltid)
        s_gid.add(line.lt.ltgid)
        s_host.add(line.host)

    print("[DB status] in {0} - {1}".format(top_dt, end_dt))
    print("Registered log lines : {0}".format(cnt_line))
    print("Log templates : {0}".format(len(s_ltid)))
    print("Log template groups : {0}".format(len(s_gid)))
    print("Hosts : {0}".format(len(s_host)))


def show_lt(conf, **kwargs):
    ld = LogData(conf)
    if "simple" in kwargs and kwargs["simple"]:
        for ltobj in ld.iter_lt():
            print(ltobj)
    else:
        print(ld.show_ltgroup_cond(**kwargs))


def dump_lt(conf):
    ld = LogData(conf)
    print(ld.show_all_lt())


def show_lt_import(conf):
    ld = LogData(conf)
    for ltobj in ld.iter_lt():
        print(" ".join(ltobj.ltw))


def show_template_table(conf):
    ld = LogData(conf)
    print(ld.show_template_table())


def dump_template_table(conf):
    ld = LogData(conf)
    print(ld.dump_template_table())


def show_all_host(conf, top_dt=None, end_dt=None):
    ld = LogData(conf)
    for host in ld.whole_host():
        print(host)


def agg_words(conf, target="all"):
    """Return dict of words in all log messages and their counts. 

    Args:
        conf
        target (str): (all, description, variable) is available

    Returns:
        dict: key = word, val = counts
    """

    def getall(d, lm):
        for w in lm.l_w:
            d[w] += 1

    def getdesc(d, lm):
        for w in lm.lt.ltw:
            if w == lm.lt.sym:
                pass
            else:
                d[w] += 1

    def getvar(d, lm):
        for w in lm.var():
            d[w] += 1

    assert target in ("all", "description", "variable")
    if target == "all":
        func = getall
    elif target == "description":
        func = getdesc
    elif target == "variable":
        func = getvar
    else:
        raise NotImplementedError

    ld = LogData(conf)
    top_dt, end_dt = ld.dt_term()
    d = defaultdict(int)
    for lm in ld.iter_lines(top_dt=top_dt, end_dt=end_dt):
        func(d, lm)

    return d


def migrate(conf):
    ld = LogData(conf, edit=True)
    ld.db._init_index()
    ld.update_area()
    ld.commit_db()


def remake_ltgroup(conf):
    ld = LogData(conf, edit=True)
    ld.init_ltmanager()
    ld.ltm.remake_ltg()
    ld.commit_db()


def reload_area(conf):
    ld = LogData(conf, edit=True)
    ld.init_ltmanager()
    ld.update_area()
    ld.commit_db()


def anonymize(conf):
    ld = LogData(conf, edit=True)
    d_cond = {}
    d_update = {"words": ""}
    ld.db.update_log(d_cond, d_update)
    ld.commit_db()


def data_from_db(conf, dirname, method, reset):
    rod = RestoreOriginalData(dirname, method=method, reset=reset)
    ld = LogData(conf)
    top_dt, end_dt = ld.whole_term()
    for lm in ld.iter_lines(top_dt=top_dt, end_dt=end_dt):
        rod.add(lm)
    rod.commit()


def data_from_data(conf, targets, dirname, method, reset):
    rod = RestoreOriginalData(dirname, method=method, reset=reset)
    lp = load_log2seq(conf)
    ha = host_alias.init_hostalias(conf)
    drop_undefhost = conf.getboolean("database", "undefined_host")
    for f in iter_files(targets):
        for line in f:
            pline = parse_line(line, lp)
            if pline is None:
                continue
            dt = pline["timestamp"]
            pline = normalize_host(line, pline, ha, None, drop_undefhost)
            if pline is None:
                continue
            rod.add_str(dt, line)
    rod.commit()



#!/usr/bin/env python

import sys
import os
from setuptools import setup


def load_readme():
    with open('README.md', 'r') as f:
        return f.read()


sys.path.append("./tests")
package_name = 'amulog'
data_dir = "/".join((package_name, "data"))
data_files = ["/".join((data_dir, fn)) for fn in os.listdir(data_dir)]

setup(name='amulog',
      version='0.0.7',
      description='A system log management tool with automatically generated log templates.',
      long_description=load_readme(),
      long_description_content_type='text/markdown',
      author='Satoru Kobayashi',
      author_email='sat@nii.ac.jp',
      url='https://github.com/cpflat/amulog/',
      install_requires=['numpy>=1.9.2',
                        'scipy>=1.2.0',
                        'scikit-learn>=0.20.2',
                        'python-dateutil>=2.8.0',
                        'log2seq>=0.0.3', ],
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: Information Technology',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: BSD License',
          'Programming Language :: Python :: 3.7',
          'Topic :: Scientific/Engineering :: Information Analysis',
          'Topic :: Software Development :: Libraries :: Python Modules'],
      license='The 3-Clause BSD License',

      packages=['amulog'],
      package_data={'amulog': data_files},
      test_suite="tests"
      )



#!/usr/bin/env python
# coding: utf-8

import os
import datetime
from collections import namedtuple

TableKey = namedtuple("TableKey", ("key", "type", "attr"))
Condition = namedtuple("Condition", ("key", "opr", "val", "repl"))
Setstate = namedtuple("SetState", ("key", "val"))


class database(object):

    # d_key in create_table : key = key_name, val = [type, attr, attr...]
    # d_opr : operand to use in compararizon
    # l_repl : values with given keys are replaced in sql_query
    #          to generate sql with subquery

    def __init__(self, dbpath):
        raise NotImplementedError

    def db_exists(self):
        raise NotImplementedError

    def reset(self):
        raise NotImplementedError

    def commit(self):
        raise NotImplementedError
        
    def strftime(self, dt):
        if isinstance(dt, datetime.datetime):
            return dt.strftime('%Y-%m-%d %H:%M:%S')
        elif isinstance(dt, str):
            return dt
        else:
            raise TypeError

    def strptime(self, string):
        return datetime.datetime.strptime(string, '%Y-%m-%d %H:%M:%S')

    def datetime(self, ret):
        raise NotImplementedError

    def _ph(self, varname):
        raise NotImplementedError
        
    def _set_state(self, l_setstate):
        if len(l_setstate) == 0:
            raise ValueError("empty setstates")
        l_buf = []
        for ss in l_setstate:
            l_buf.append("{0} = {1}".format(ss.key, self._ph(ss.val)))
        return ", ".join(l_buf)

    def _cond_state(self, l_cond):
        l_buf = []
        for cond in l_cond:
            if cond.repl:
                buf = "{0.key} {0.opr} {1}".format(cond, self._ph(cond.val))
            else:
                buf = "{0.key} {0.opr} ({0.val})".format(cond)
            l_buf.append(buf)
        return " and ".join(l_buf)

    def _table_key_type(self, type_str):
        # allowed typename : integer, text, datetime
        raise NotImplementedError

    def _table_key_attr(self, attr):
        # allowed attr : primary_key, auto_increment, not_null
        raise NotImplementedError

    def _index_key(self, tablekey):
        raise NotImplementedError

    def join_sql(self, join_opt, table_name1, table_name2, key1, key2):
        return "{1} {0} join {2} on {1}.{3} = {2}.{4}".format(join_opt,
                table_name1, table_name2, key1, key2)

    def create_table_sql(self, table_name, l_key):
        l_def = []
        for key in l_key:
            type_name = self._table_key_type(key.type)
            if len(key.attr) > 0:
                l_attr = []
                for a in key.attr:
                    l_attr.append(self._table_key_attr(a))
                l_def.append("{0} {1} {2}".format(key.key, type_name,
                        " ".join(l_attr)))
            else:
                l_def.append("{0} {1}".format(key.key, type_name))
        sql = "create table {0} ({1})".format(table_name, ", ".join(l_def))
        return sql

    def create_index_sql(self, table_name, index_name, l_key):
        sql = "create index {0} on {1}({2})".format(index_name,
                table_name, ", ".join([self._index_key(key) for key in l_key]))
        return sql

    def select_sql(self, table_name, l_key, l_cond = [], opt = []):
        # now only "distinct" is allowed for opt
        sql_header = "select"
        if "distinct" in opt:
            sql_header += " distinct"
        sql = "{0} {1} from {2}".format(sql_header, ", ".join(l_key),
                table_name)
        if len(l_cond) > 0:
            sql += " where {0}".format(self._cond_state(l_cond))
        return sql 

    def insert_sql(self, table_name, l_setstate):
        l_key, l_val = zip(*[(ss.key, self._ph(ss.val)) for ss in l_setstate])
        sql = "insert into {0} ({1}) values ({2})".format(table_name,
                ", ".join(l_key), ", ".join(l_val))
        return sql

    def update_sql(self, table_name, l_setstate, l_cond = []):
        sql = "update {0} set {1}".format(table_name,
                self._set_state(l_setstate))
        if len(l_cond) > 0:
            sql += " where {0}".format(self._cond_state(l_cond))
        return sql 

    def delete_sql(self, table_name, l_cond = []):
        sql = "delete from {0}".format(table_name)
        if len(l_cond) > 0:
            sql += " where {0}".format(self._cond_state(l_cond))
        return sql

    def drop_sql(self, table_name):
        return "drop table {0}".format(table_name)

    def execute(self, sql, args):
        raise NotImplementedError

    def get_table_names(self):
        raise NotImplementedError


class sqlite3(database):

    def __init__(self, dbpath):
        self.dbpath = dbpath
        self.connect = None

    def __del__(self):
        if self.connect is not None:
            self.connect.commit()
            self.connect.close()
    
    def _open(self):
        import sqlite3 as sqlite3_mod
        self.connect = sqlite3_mod.connect(self.dbpath)
        self.connect.text_factory = str
    
    def db_exists(self):
        if os.path.exists(self.dbpath):
            return True
        else:
            return False

    def reset(self):
        if os.path.exists(self.dbpath):
            os.remove(self.dbpath)

    def commit(self):
        if self.connect is not None:
            self.connect.commit()

    def datetime(self, ret):
        return self.strptime(ret)

    def _ph(self, varname):
        return ":{0}".format(varname)

    def _table_key_type(self, type_str):
        if type_str == "datetime":
            return "text"
        else:
            return type_str

    def _table_key_attr(self, attr):
        if attr == "primary_key":
            return "primary key"
        elif attr == "auto_increment":
            return "autoincrement"
        elif attr == "not_null":
            return "not null"
        else:
            raise NotImplementedError

    def _index_key(self, tablekey):
        return tablekey.key
        
    def execute(self, sql, args = {}):
        #print sql
        #if len(args) > 0: print args
        if self.connect is None:
            self._open()
        cursor = self.connect.cursor()
        if len(args) == 0:
            cursor.execute(sql)
        else:
            cursor.execute(sql, args)
        return cursor

    def executemany(self, sql, iter_args):
        if self.connect is None:
            self._open()
        cursor = self.connect.cursor()
        cursor.executemany(sql, iter_args)
        return cursor

    def get_table_names(self):
        sql = "select name from sqlite_master"
        cursor = self.execute(sql)
        return [row[0] for row in cursor]


class mysql(database):

    def __init__(self, host, dbname, user, passwd):
        global MySQLdb
        import MySQLdb
        self.host = host
        self.dbname = dbname
        self.user = user
        self.passwd = passwd
        self.connect = None

    def __del__(self):
        if self.connect is not None:
            self.connect.commit()
            self.connect.close()

    def _open(self):
        if not self.db_exists():
            self._init_database()
        self.connect = MySQLdb.connect(host = self.host, db = self.dbname,
                                       user = self.user, passwd = self.passwd)

    def _init_database(self):
        connect = self._connect_root()
        cursor = connect.cursor()
        cursor.execute("create database {0}".format(self.dbname))

    def _connect_root(self):
        return MySQLdb.connect(host = self.host, user = self.user,
                               passwd = self.passwd)

    def db_exists(self):
        connect = self._connect_root()
        cursor = connect.cursor()
        cursor.execute("show databases")
        return self.dbname in [row[0] for row in cursor]

    def reset(self):
        if self.db_exists():
            connect = self._connect_root()
            cursor = connect.cursor()
            cursor.execute("drop database {0}".format(self.dbname))
            connect.commit()

    def commit(self):
        if self.connect is not None:
            self.connect.commit()

    def datetime(self, ret):
        return ret

    def _ph(self, varname):
        return "%({0})s".format(varname)

    def _table_key_type(self, type_str):
        if type_str == "integer":
            return "int"
        else:
            return type_str

    def _table_key_attr(self, attr):
        if attr == "primary_key":
            return "primary key"
        elif attr == "auto_increment":
            # differ from sqlite
            return "auto_increment"
        elif attr == "not_null":
            return "not null"
        else:
            raise NotImplementedError
    
    def _index_key(self, tablekey):
        if tablekey.type == "text":
            return "{0}({1})".format(tablekey.key, tablekey.attr[0])
        else:
            return tablekey.key

    def execute(self, sql, args = {}):
        #print sql
        #if len(args) > 0: print args
        if self.connect is None:
            self._open()
        cursor = self.connect.cursor()
        if len(args) == 0:
            cursor.execute(sql)
        else:
            cursor.execute(sql, args)
        return cursor

    def executemany(self, sql, iter_args):
        if self.connect is None:
            self._open()
        cursor = self.connect.cursor()
        cursor.executemany(sql, iter_args)
        return cursor

    def get_table_names(self):
        sql = "show tables"
        cursor = self.execute(sql)
        return [row[0] for row in cursor]


def tablekey(keyname, typename, attributes = tuple()):
    return TableKey(keyname, typename, tuple(attributes))


def cond(keyname, operand, value, replace = True):
    # value : variable name (key of args) if replace = True
    #         subquery statement if replace = False
    return Condition(keyname, operand, value, replace)


def setstate(keyname, value):
    return Setstate(keyname, value)





#!/usr/bin/env python
# coding: utf-8

import os
import re
import pickle
import logging
from collections import defaultdict
from importlib import import_module

from . import config
from . import strutil

REPLACER = "**"
REPLACER_HEAD = "*"
REPLACER_TAIL = "*"
REPLACER_REGEX = re.compile(r"\*[A-Z]*?\*")  # shortest match

_logger = logging.getLogger(__package__)


class LTManager(object):
    """
    A log template manager. This class define log templates from messages.
    In addition, this class update log template table on memory and DB.

    Log template generation process can be classified to following 2 types.
        Grouping : Classify messages into groups, and generate template from
                the common parts of group members.
        Templating : Estimate log template from messages by classifying words,
                and make groups that have common template.

    Attributes:
        #TODO
    
    """

    # adding lt to db (ltgen do not add)

    def __init__(self, conf, db, lttable, reset_db):
        self._reset_db = reset_db
        self.filename = conf.get("log_template", "indata_filename")

        self._db = db
        self._lttable = lttable
        self._table = TemplateTable()
        self._ltgen = None
        self._ltspl = None
        self._ltgroup = None

    def set_ltgen(self, ltgen):
        self._ltgen = ltgen

    def set_ltgroup(self, ltgroup):
        self._ltgroup = ltgroup
        if not self._reset_db:
            self._ltgroup.restore_ltg(self._db, self._lttable)

    def set_ltspl(self, ltspl):
        self._ltspl = ltspl

    def process_offline(self, plines):
        d_tid = self._ltgen.process_offline(plines)

        for mid, pline in enumerate(plines):
            l_w = pline["words"]
            l_s = pline["symbols"]
            tid = d_tid[mid]
            tpl = self._table[tid]
            ltw = self._ltspl.replace_variable(l_w, tpl, REPLACER)
            ltid = self._ltspl.search(tid, ltw)
            if ltid is None:
                ltline = self.add_lt(ltw, l_s)
                self._table.addcand(tid, ltline.ltid)
            else:
                self.count_lt(ltid)
                ltline = self._lttable[ltid]
            yield ltline

    def process_line(self, pline):

        def _lt_diff(ltid, ltw):
            d_diff = {}
            for wid, new_w, old_w in zip(range(len(ltw)), ltw,
                                         self._lttable[ltid].ltw):
                if new_w == old_w:
                    pass
                else:
                    d_diff[wid] = new_w
            return d_diff

        def _lt_repl(ltw, d_diff):
            ret = []
            for wid, w in enumerate(ltw):
                if wid in d_diff:
                    ret.append(d_diff[wid])
                else:
                    ret.append(w)
            return ret

        l_w = pline["words"]
        l_s = pline["symbols"]
        tid, state = self._ltgen.process_line(pline)
        if tid is None:
            return None

        tpl = self._table[tid]
        ltw = self._ltspl.replace_variable(l_w, tpl, REPLACER)
        if state == LTGen.state_added:
            ltline = self.add_lt(ltw, l_s)
            self._table.addcand(tid, ltline.ltid)
        else:
            ltid = self._ltspl.search(tid, ltw)
            if ltid is None:
                # tpl exists, but no lt matches
                ltline = self.add_lt(ltw, l_s)
                self._table.addcand(tid, ltline.ltid)
            else:
                if state == LTGen.state_changed:
                    # update all lt that belong to the edited tpl
                    d_diff = _lt_diff(ltid, ltw)
                    for temp_ltid in self._table.getcand(tid):
                        if temp_ltid == ltid:
                            self.replace_and_count_lt(ltid, ltw)
                        else:
                            old_ltw = self._lttable[temp_ltid]
                            new_ltw = _lt_repl(old_ltw, d_diff)
                            self.replace_lt(ltid, new_ltw)
                elif state == LTGen.state_unchanged:
                    self.count_lt(ltid)
                else:
                    raise AssertionError
                ltline = self._lttable[ltid]

        return ltline

    def add_lt(self, l_w, l_s, cnt=1):
        # add new lt to db and table
        ltid = self._lttable.next_ltid()
        ltline = LogTemplate(ltid, None, l_w, l_s, cnt)
        ltgid = self._ltgroup.add(ltline)
        ltline.ltgid = ltgid
        self._lttable.add_lt(ltline)
        self._db.add_lt(ltline)
        return ltline

    def replace_lt(self, ltid, l_w, l_s=None, cnt=None):
        self._lttable[ltid].replace(l_w, l_s, cnt)
        self._db.update_lt(ltid, l_w, l_s, cnt)

    def replace_and_count_lt(self, ltid, l_w, l_s=None):
        cnt = self._lttable[ltid].add()
        self._lttable[ltid].replace(l_w, l_s, None)
        self._db.update_lt(ltid, l_w, l_s, cnt)

    def count_lt(self, ltid):
        cnt = self._lttable[ltid].count()
        self._db.update_lt(ltid, None, None, cnt)

    def remove_lt(self, ltid):
        self._lttable.remove_lt(ltid)
        self._db.remove_lt(ltid)

    def remake_ltg(self):
        self._db.reset_ltg()
        self._ltgroup._init_dict()
        temp_lttable = self._lttable
        self._ltgroup._lttable = LTTable()

        for ltline in temp_lttable:
            ltgid = self._ltgroup.add(ltline)
            ltline.ltgid = ltgid
            self._ltgroup._lttable.add_lt(ltline)
            self._db.add_ltg(ltline.ltid, ltgid)
        assert self._ltgroup._lttable.ltdict == temp_lttable.ltdict

    def load(self):
        with open(self.filename, 'rb') as f:
            obj = pickle.load(f)
        table_data, ltgen_data, ltgroup_data = obj
        self._table.load(table_data)
        self._ltgen.load(ltgen_data)
        self._ltgroup.load(ltgroup_data)

    def dump(self):
        table_data = self._table.dumpobj()
        ltgen_data = self._ltgen.dumpobj()
        ltgroup_data = self._ltgroup.dumpobj()
        obj = (table_data, ltgen_data, ltgroup_data)
        with open(self.filename, 'wb') as f:
            pickle.dump(obj, f)


class LTTable:

    def __init__(self):
        self.ltdict = {}

    def __iter__(self):
        return self._generator()

    def _generator(self):
        for ltid in self.ltdict:
            yield self.ltdict[ltid]

    def __len__(self):
        return len(self.ltdict)

    def __getitem__(self, key):
        assert isinstance(key, int)
        if key not in self.ltdict:
            raise IndexError("index out of range")
        return self.ltdict[key]

    def next_ltid(self):
        cnt = 0
        while cnt in self.ltdict:
            cnt += 1
        else:
            return cnt

    def restore_lt(self, ltid, ltgid, ltw, lts, count):
        assert ltid not in self.ltdict
        self.ltdict[ltid] = LogTemplate(ltid, ltgid, ltw, lts, count)

    def add_lt(self, ltline):
        assert not ltline.ltid in self.ltdict
        self.ltdict[ltline.ltid] = ltline

    def remove_lt(self, ltid):
        self.ltdict.pop(ltid)


class LogTemplate:

    def __init__(self, ltid, ltgid, ltw, lts, count):
        if len(ltw) == 0:
            raise ValueError("empty ltw, failed to generate LogTemplate")
        self.ltid = ltid
        self.ltgid = ltgid
        self.ltw = ltw
        self.lts = lts
        self.cnt = count

    def __iter__(self):
        return self.ltw

    def __str__(self):
        return self.restore_message(self.ltw)

    def get(self, key):
        if key == "ltid":
            return self.ltid
        elif key == "ltgid":
            return self.ltgid
        else:
            raise KeyError

    def var(self, l_w):
        if len(l_w) == 0:
            return [REPLACER for w in self.ltw if w == REPLACER]
        else:
            return [w_org for w_org, w_lt in zip(l_w, self.ltw)
                    if w_lt == REPLACER]

    def var_location(self):
        return [i for i, w_lt in enumerate(self.ltw) if w_lt == REPLACER]

    def restore_message(self, l_w, esc=False):
        if l_w is None or len(l_w) == 0:
            l_w = self.ltw
        if esc:
            l_w = [w for w in l_w]
        else:
            l_w = [strutil.restore_esc(w) for w in l_w]

        if self.lts is None:
            return "".join(l_w)
        else:
            return "".join([s + w for w, s in zip(l_w + [""], self.lts)])

    def count(self):
        self.cnt += 1
        return self.cnt

    def replace(self, l_w, l_s=None, count=None):
        self.ltw = l_w
        if l_s is not None:
            self.lts = l_s
        if count is not None:
            self.cnt = count


class TemplateTable:
    """Temporal template table for log template generator."""

    def __init__(self):
        self._d_tpl = {}  # key = tid, val = template
        self._d_rtpl = {}  # key = key_template, val = tid
        self._d_cand = defaultdict(list)  # key = tid, val = List[ltid]
        self._last_modified = None  # used for LTGenJoint

    def __str__(self):
        ret = []
        for tid, tpl in self._d_tpl.items():
            ret.append(" ".join([str(tid)] + tpl))
        return "\n".join(ret)

    def __iter__(self):
        return self._generator()

    def _generator(self):
        for tid in self._d_tpl:
            yield self._d_tpl[tid]

    def __getitem__(self, key):
        assert isinstance(key, int)
        if key not in self._d_tpl:
            raise IndexError("index out of range")
        return self._d_tpl[key]

    def __len__(self):
        return len(self._d_tpl)

    def next_tid(self):
        cnt = 0
        while cnt in self._d_tpl:
            cnt += 1
        else:
            return cnt

    def tids(self):
        return self._d_tpl.keys()

    @staticmethod
    def _key_template(template):
        l_word = [strutil.add_esc(w) for w in template]
        return "@".join(l_word)

    def exists(self, template):
        key = self._key_template(template)
        return key in self._d_rtpl

    def get_tid(self, template):
        key = self._key_template(template)
        return self._d_rtpl[key]

    def get_template(self, tid):
        return self._d_tpl[tid]

    def add(self, template):
        tid = self.next_tid()
        self._d_tpl[tid] = template
        self._d_rtpl[self._key_template(template)] = tid
        return tid

    def replace(self, tid, template):
        self._last_modified = self._d_tpl[tid]
        self._d_tpl[tid] = template
        self._d_rtpl[self._key_template(template)] = tid

    def get_updated(self):
        return self._last_modified

    def getcand(self, tid):
        return self._d_cand[tid]

    def addcand(self, tid, ltid):
        self._d_cand[tid].append(ltid)

    def load(self, obj):
        self._d_tpl, self._d_cand = obj
        for tid, tpl in self._d_tpl.items():
            self._d_rtpl[self._key_template(tpl)] = tid

    def dumpobj(self):
        return self._d_tpl, self._d_cand


class LTGen(object):
    stateful = True
    state_added = 0
    state_changed = 1
    state_unchanged = 2

    def __init__(self, table: TemplateTable):
        self._table = table

    def get_tpl(self, tid):
        return self._table[tid]

    def add_tpl(self, ltw, tid=None):
        if tid is None:
            tid = self._table.add(ltw)
        return tid

    def update_tpl(self, ltw, tid):
        self._table.replace(tid, ltw)

    def merge_tpl(self, l_w, tid):
        old_tpl = self._table[tid]
        new_tpl = merged_template(old_tpl, l_w)
        if old_tpl == new_tpl:
            return self.state_unchanged
        else:
            self.update_tpl(new_tpl, tid)
            return self.state_changed

    def update_table(self, tpl):
        if tpl is None:
            return None, None
        elif self._table.exists(tpl):
            tid = self._table.get_tid(tpl)
            return tid, self.state_unchanged
        else:
            tid = self._table.add(tpl)
            return tid, self.state_added

    def preprocess(self, plines):
        """This function do the pre-process for given data if needed.
        This function is called by process_init_data."""
        pass

    def process_offline(self, plines):
        """If there is no need of special process for init phase,
        this function simply call process_line multiple times.
        """
        self.preprocess(plines)
        d = {}
        for mid, pline in enumerate(plines):
            tid, state = self.process_line(pline)
            d[mid] = tid
        return d

    def process_line(self, pline):
        """Estimate log template for given message.
        This method works in incremental processing phase.

        Args:
            pline (dict): parsed log message with log2seq

        Returns:
            tid (int): A template id in TemplateTable.
            state (int)
        """
        raise NotImplementedError

    def postprocess(self, plines):
        """This function do the post-process for given data if needed.
        This function is called by process_init_data."""
        pass

    def load(self, loadobj):
        raise NotImplementedError

    def dumpobj(self):
        raise NotImplementedError


class LTGenStateless(LTGen):

    """Subclasses of LTGenStateless is acceptable
    for multiprocessing in offline template generation."""

    stateful = False

    def generate_tpl(self, pline):
        raise NotImplementedError

    def process_line(self, pline):
        tpl = self.generate_tpl(pline)
        return self.update_table(tpl)

    def load(self, loadobj):
        pass

    def dumpobj(self):
        return None


class LTGenJoint(LTGen):

    def __init__(self, table: TemplateTable, l_ltgen, ltgen_import_index=None):
        super().__init__(table)
        self._l_ltgen = l_ltgen
        self._import_index = ltgen_import_index

    def _update_ltmap(self, pline, index, tid, state):
        from .lt_import import LTGenImport
        if (self._import_index is not None) and \
                (not index == self._import_index):
            ltgen_import: LTGenImport = self._l_ltgen[self._import_index]
            if state == self.state_added:
                ltgen_import.add_definition(pline["words"])
            elif state == self.state_changed:
                old_tpl = self._table.get_updated()
                new_tpl = self._table.get_template(tid)
                ltgen_import.update_definition(old_tpl, new_tpl)

    def process_line(self, pline):
        for index, ltgen in enumerate(self._l_ltgen):
            tid, state = ltgen.process_line(pline)
            if tid is not None:
                self._update_ltmap(pline, index, tid, state)
                return tid, state
        else:
            msg = "Template for a message not matched/generated: {0}".format(
                pline["message"])
            _logger.debug(msg)
            return None, None

    def load(self, loadobj):
        for ltgen, ltgen_data in zip(self._l_ltgen, loadobj):
            ltgen.load(ltgen_data)

    def dumpobj(self):
        return [ltgen.dumpobj() for ltgen in self._l_ltgen]


class LTGenMultiProcess(LTGen):

    _ltgen: LTGenStateless

    def __init__(self, table: TemplateTable, n_proc, kwargs):
        super().__init__(table)
        self._n_proc = n_proc
        self._ltgen = init_ltgen(**kwargs)
        assert not self._ltgen.stateful, \
            "multiprocessing is limited to stateless methods"

        from multiprocessing import Pool
        self._pool = Pool(processes=self._n_proc)

    def load(self, _):
        pass

    def dumpobj(self):
        return None

    @staticmethod
    def _task(args):
        ltgen, message_id, pline = args
        template = ltgen.generate_tpl(pline)
        return message_id, template

    def process_line(self, _):
        raise ValueError("use process_offline")

    def process_offline(self, plines):
        l_args = [(self._ltgen, mid, pline)
                  for mid, pline in enumerate(plines)]
        try:
            d_tpl = {}
            for mid, tpl in self._pool.imap_unordered(self._task, l_args):
                d_tpl[mid] = tpl
            self._pool.close()
        except KeyboardInterrupt:
            self._pool.terminate()
            exit()
        else:
            ret = {}
            for mid, tpl in d_tpl.items():
                tid, _ = self.update_table(tpl)
                ret[mid] = tid
            return ret


class LTGroup(object):

    # usually used as super class of other ltgroup
    # If used directly, this class will work as a dummy
    # (ltgid is always same as ltid)

    def __init__(self):
        self._init_dict()

    def _init_dict(self):
        self._d_group = {}  # key : groupid, val : [ltline, ...]
        self._d_rgroup = {}  # key : ltid, val : groupid

    def _next_groupid(self):
        cnt = 0
        while cnt in self._d_group:
            cnt += 1
        else:
            return cnt

    def add(self, ltline):
        gid = ltline.ltid
        self.add_ltid(gid, ltline)
        return gid

    def add_ltid(self, gid, ltline):
        self._d_group.setdefault(gid, []).append(ltline)
        self._d_rgroup[ltline.ltid] = gid

    def restore_ltg(self, db, table):
        for ltid, ltgid in db.iter_ltg_def():
            self._d_group.setdefault(ltgid, []).append(table[ltid])
            self._d_rgroup[ltid] = ltgid

    def load(self, loadobj):
        pass

    def dumpobj(self):
        return None


class LTPostProcess(object):

    def __init__(self, conf, table, lttable, l_alg):
        self._table = table
        self._lttable = lttable
        self._rules = []
        for alg in l_alg:
            if alg == "dummy":
                self._rules.append(VariableLabelRule())
            elif alg == "host":
                self._rules.append(VariableLabelHost(conf))
            else:
                raise NotImplementedError

        self.sym_header = REPLACER_HEAD
        self.sym_footer = REPLACER_TAIL
        #self.sym_header = conf.get("log_template",
        #                           "labeled_variable_symbol_header")
        #self.sym_footer = conf.get("log_template",
        #                           "labeled_variable_symbol_footer")

    def _labeled_variable(self, w):
        return "".join((self.sym_header, w, self.sym_footer))

    def replace_variable(self, l_w, tpl, sym):
        ret = []
        for org_w, tpl_w in zip(l_w, tpl):
            if tpl_w == sym:
                for r in self._rules:
                    ww = r.replace_word(org_w)
                    if ww is not None:
                        ret.append(self._labeled_variable(ww))
                        break
                else:
                    ret.append(tpl_w)
            else:
                ret.append(tpl_w)
        return ret

    def search(self, tid, ltw):
        """Search existing candidates of template derivation. Return None
        if no possible candidates found."""
        l_ltid = self._table.getcand(tid)
        for ltid in l_ltid:
            if self._lttable[ltid].ltw == ltw:
                return ltid
        else:
            return None

        # if len(self._table.getcand(tid)) == 0:
        #    return None
        # else:
        #    return self._table.getcand(tid)[0]


class VariableLabelRule(object):

    def __init__(self):
        pass

    def replace_word(self, w):
        """str: If the given word is a member of some host group,
        return the group name. Otherwise, return None."""
        return None


class VariableLabelHost(VariableLabelRule):

    def __init__(self, conf):
        super().__init__()
        from . import host_alias
        self.ha = host_alias.init_hostalias(conf)

    def replace_word(self, w):
        return self.ha.get_group(w)


def init_ltgen(conf, table, method, shuffle=False):
    kwargs = {"conf": conf,
              "table": table,
              "shuffle": shuffle}
    if method == "import":
        from . import lt_import
        return lt_import.init_ltgen_import(**kwargs)
    elif method == "import-ext":
        from . import lt_import_ext
        return lt_import_ext.init_ltgen_import_ext(**kwargs)
    elif method == "crf":
        from amulog.alg.crf import lt_crf
        return lt_crf.init_ltgen_crf(**kwargs)
    elif method == "re":
        from amulog import lt_regex
        return lt_regex.init_ltgen_regex(**kwargs)
    elif method == "va":
        from . import lt_va
        return lt_va.init_ltgen_va(**kwargs)
    else:
        modname = "amulog.alg." + method
        alg_module = import_module(modname)
        return alg_module.init_ltgen(**kwargs)


def init_ltgen_methods(conf, table, lt_methods=None, shuffle=False,
                       multiprocess=None):
    if lt_methods is None:
        lt_methods = config.getlist(conf, "log_template", "lt_methods")
    if multiprocess is None:
        multiprocess = conf.getboolean("log_template", "ltgen_multiprocess")
    ltgen_kwargs = {"conf": conf,
                    "table": table,
                    "shuffle": shuffle}

    if multiprocess:
        n_proc = conf["log_template"]["n_proc"].strip()
        if n_proc.isdigit():
            n_proc = int(n_proc)
        else:
            n_proc = None
        assert len(lt_methods) == 1, \
            "for multiprocessing, lt_methods should be single"
        ltgen_kwargs["method"] = lt_methods[0]
        return LTGenMultiProcess(table, n_proc, ltgen_kwargs)
    elif len(lt_methods) > 1:
        l_ltgen = []
        import_index = None
        for mid, method_name in enumerate(lt_methods):
            l_ltgen.append(init_ltgen(conf, table, method_name, shuffle))
            if method_name == "import":
                import_index = mid
        return LTGenJoint(table, l_ltgen, import_index)
    elif len(lt_methods) == 1:
        return init_ltgen(conf, table, lt_methods[0], shuffle)
    else:
        raise ValueError


def init_ltmanager(conf, db, table, reset_db):
    """Initializing ltmanager by loading argument parameters."""
    # sym = conf.get("log_template", "variable_symbol")
    ltm = LTManager(conf, db, table, reset_db)

    ltm.set_ltgen(init_ltgen_methods(conf, ltm._table))

    ltg_alg = conf.get("log_template", "ltgroup_alg")
    if ltg_alg == "shiso":
        from amulog.alg.shiso import shiso
        ltgroup = shiso.LTGroupSHISO(table,
                                     ngram_length=conf.getint(
                                            "log_template_shiso", "ltgroup_ngram_length"),
                                     th_lookup=conf.getfloat(
                                            "log_template_shiso", "ltgroup_th_lookup"),
                                     th_distance=conf.getfloat(
                                            "log_template_shiso", "ltgroup_th_distance"),
                                     mem_ngram=conf.getboolean(
                                            "log_template_shiso", "ltgroup_mem_ngram")
                                     )
    elif ltg_alg == "ssdeep":
        from . import lt_misc
        ltgroup = lt_misc.LTGroupFuzzyHash(table)
    elif ltg_alg == "none":
        ltgroup = LTGroup()
    else:
        raise ValueError("ltgroup_alg({0}) invalid".format(ltg_alg))
    ltm.set_ltgroup(ltgroup)

    post_alg = config.gettuple(conf, "log_template", "post_alg")
    ltspl = LTPostProcess(conf, ltm._table, ltm._lttable, post_alg)
    ltm.set_ltspl(ltspl)

    if os.path.exists(ltm.filename) and not reset_db:
        ltm.load()

    return ltm


def merged_template(m1, m2):
    """Return common area of log message (to be log template)"""
    ret = []
    for w1, w2 in zip(m1, m2):
        if w1 == w2:
            ret.append(w1)
        else:
            ret.append(REPLACER)
    return ret
